"""This module contains several class definitions for parts of the
PEST control file: ControlData, RegData and SvdData.  These
classes are automatically created and appended to a Pst object;
users shouldn't need to deal with these classes explicitly
"""
from __future__ import print_function, division
import os
import copy
import warnings
import numpy as np
import pandas
from ..pyemu_warnings import PyemuWarning
pandas.options.display.max_colwidth = 100
#from pyemu.pst.pst_handler import SFMT,SFMT_LONG,FFMT,IFMT

#formatters
SFMT = lambda x: "{0:>20s}".format(str(x))
SFMT_LONG = lambda x: "{0:>50s}".format(str(x))
IFMT = lambda x: "{0:>10d}".format(int(x))
FFMT = lambda x: "{0:>15.6E}".format(float(x))

CONTROL_DEFAULT_LINES = """restart estimation
     0     0       0    0      0  0
0  0  single  point  1  0  0 noobsreref
2.000000e+001  -3.000000e+000  3.000000e-001  1.000000e-002 -7 999 lamforgive noderforgive
1.000000e+001  1.000000e+001  1.000000e-003  0  0
1.000000e-001 1 1.1 noaui nosenreuse noboundscale
30 1.000000e-002  3  3  1.000000e-002  3  0.0 1  -1.0
0  0  0  0 jcosave verboserec jcosaveitn reisaveitn parsaveitn noparsaverun"""\
    .lower().split('\n')

CONTROL_VARIABLE_LINES = """RSTFLE PESTMODE
NPAR NOBS NPARGP NPRIOR NOBSGP [MAXCOMPDIM]
NTPLFLE NINSFLE PRECIS DPOINT [NUMCOM] [JACFILE] [MESSFILE] [OBSREREF]
RLAMBDA1 RLAMFAC PHIRATSUF PHIREDLAM NUMLAM [JACUPDATE] [LAMFORGIVE] [DERFORGIVE]
RELPARMAX FACPARMAX FACORIG [IBOUNDSTICK] [UPVECBEND]
PHIREDSWH [NOPTSWITCH] [SPLITSWH] [DOAUI] [DOSENREUSE] [BOUNDSCALE]
NOPTMAX PHIREDSTP NPHISTP NPHINORED RELPARSTP NRELPAR [PHISTOPTHRESH] [LASTRUN] [PHIABANDON]
ICOV ICOR IEIG [IRES] [JCOSAVE] [VERBOSEREC] [JCOSAVEITN] [REISAVEITN] [PARSAVEITN] [PARSAVERUN]"""\
    .lower().split('\n')

REG_VARIABLE_LINES = """PHIMLIM PHIMACCEPT [FRACPHIM] [MEMSAVE]
WFINIT WFMIN WFMAX [LINREG] [REGCONTINUE]
WFFAC WFTOL IREGADJ [NOPTREGADJ REGWEIGHTRAT [REGSINGTHRESH]]""".lower().split('\n')

REG_DEFAULT_LINES = """   1.0e-10    1.05e-10  0.1  nomemsave
1.0 1.0e-10 1.0e10 linreg continue
1.3  1.0e-2  1 1.5 1.5 0.5""".lower().split('\n')

class RegData(object):
    """ an object that encapsulates the regularization section
    of the PEST control file
    """
    def __init__(self):
        self.optional_dict = {}
        for vline,dline in zip(REG_VARIABLE_LINES,REG_DEFAULT_LINES):
            vraw = vline.split()
            draw = dline.split()
            for v,d in zip(vraw,draw):
                o = False
                if '[' in v:
                    o = True
                v = v.replace('[','').replace(']','')
                super(RegData,self).__setattr__(v,d)
                self.optional_dict[v] = o
        self.should_write = ["phimlim","phimaccept","fracphim"]

    def write(self,f):
        """ write the regularization section to an open
        file handle

        Parameters
        ----------
        f : file handle

        """
        f.write("* regularization\n")
        for vline in REG_VARIABLE_LINES:
            vraw = vline.strip().split()
            for v in vraw:
                v = v.replace("[",'').replace("]",'')
                if v not in self.optional_dict.keys():
                    raise Exception("RegData missing attribute {0}".format(v))
                f.write("{0} ".format(self.__getattribute__(v)))
            f.write("\n")


    def write_keyword(self,f):
        for vline in REG_VARIABLE_LINES:
            vraw = vline.strip().split()
            for v in vraw:
                v = v.replace("[",'').replace("]",'')
                if v not in self.should_write:
                    continue
                if v not in self.optional_dict.keys():
                    raise Exception("RegData missing attribute {0}".format(v))
                f.write("{0:30} {1:>10}\n".format(v,self.__getattribute__(v)))


class SvdData(object):
    """ an object that encapsulates the singular value decomposition
    section of the PEST control file
    """
    def __init__(self,**kwargs):
        self.svdmode = kwargs.pop("svdmode",1)
        self.maxsing = kwargs.pop("maxsing",10000000)
        self.eigthresh = kwargs.pop("eigthresh",1.0e-6)
        self.eigwrite = kwargs.pop("eigwrite",1)


    def write_keyword(self,f):
        f.write("{0:30} {1:>10}\n".format("svdmode",self.svdmode))
        f.write("{0:30} {1:>10}\n".format("maxsing",self.maxsing))
        f.write("{0:30} {1:>10}\n".format("eigthresh",self.eigthresh))
        f.write("{0:30} {1:>10}\n".format("eigwrite",self.eigwrite))

    def write(self,f):
        """ write an SVD section to a file handle

        Parameters
        ----------
        f : file handle

        """
        f.write("* singular value decomposition\n")
        f.write(IFMT(self.svdmode)+'\n')
        f.write(IFMT(self.maxsing)+' '+FFMT(self.eigthresh)+"\n")
        f.write('{0}\n'.format(self.eigwrite))

    def parse_values_from_lines(self,lines):
        """ parse values from lines of the SVD section

        Parameters
        ----------
        lines : list

        """
        assert len(lines) == 3,"SvdData.parse_values_from_lines: expected " + \
                               "3 lines, not {0}".format(len(lines))
        try:
            self.svdmode = int(lines[0].strip().split()[0])
        except Exception as e:
            raise Exception("SvdData.parse_values_from_lines: error parsing" + \
                            " svdmode from line {0}: {1} \n".format(lines[0],str(e)))
        try:
            raw = lines[1].strip().split()
            self.maxsing = int(raw[0])
            self.eigthresh = float(raw[1])
        except Exception as e:
            raise Exception("SvdData.parse_values_from_lines: error parsing" + \
                            " maxsing and eigthresh from line {0}: {1} \n"\
                            .format(lines[1],str(e)))
        # try:
        #     self.eigwrite = int(lines[2].strip())
        # except Exception as e:
        #     raise Exception("SvdData.parse_values_from_lines: error parsing" + \
        #                     " eigwrite from line {0}: {1} \n".format(lines[2],str(e)))
        self.eigwrite = lines[2].strip()


class ControlData(object):
    """ an object that encapsulates the control data section
     of the PEST control file
    """
    def __init__(self):

        super(ControlData,self).__setattr__("formatters",{np.int32:IFMT,np.float64:FFMT,str:SFMT})
        super(ControlData,self).__setattr__("_df",self.get_dataframe())

        # acceptable values for most optional string inputs
        super(ControlData,self).__setattr__("accept_values",{'doaui':['aui','noaui'],
                                                              'dosenreuse':['senreuse','nosenreuse'],
                                                              'boundscale':['boundscale','noboundscale'],
                                                              'jcosave':['jcosave','nojcosave'],
                                                              'verboserec':['verboserec','noverboserec'],
                                                              'jcosaveitn':['jcosaveitn','nojcosvaeitn'],
                                                              'reisaveitn':['reisaveitn','noreisaveitn'],
                                                              'parsaveitn':['parsaveitn','noparsaveitn'],
                                                              'parsaverun':['parsaverun','noparsaverun']})

        self._df.index = self._df.name.apply(lambda x:x.replace('[',''))\
            .apply(lambda x: x.replace(']',''))
        super(ControlData, self).__setattr__("keyword_accessed", ["pestmode","noptmax"])
        counters = ["npar","nobs","npargp","nobsgp","nprior","ntplfle","ninsfle"]
        super(ControlData, self).__setattr__("counters", counters)
        #self.keyword_accessed = ["pestmode","noptmax"]

    def __setattr__(self, key, value):
        if key == "_df":
            super(ControlData,self).__setattr__("_df",value)
            return
        assert key in self._df.index, str(key)+" not found in attributes"
        self._df.loc[key,"value"] = self._df.loc[key,"type"](value)
        #super(ControlData, self).__getattr__("keyword_accessed").append(key)
        if key not in self.counters:
            self.keyword_accessed.append(key)

    def __getattr__(self, item):
        if item == "_df":
            return self._df.copy()
        assert item in self._df.index, str(item)+" not found in attributes"
        return self._df.loc[item,"value"]

    @staticmethod
    def get_dataframe():
        """ get a generic (default) control section dataframe
        
        Returns
        -------
        pandas.DataFrame : pandas.DataFrame

        """
        names = []
        [names.extend(line.split()) for line in CONTROL_VARIABLE_LINES]

        defaults = []
        [defaults.extend(line.split()) for line in CONTROL_DEFAULT_LINES]

        types, required,cast_defaults,formats = [],[],[],[]
        for name,default in zip(names,defaults):
            if '[' in name or ']' in name:
                required.append(False)
            else:
                required.append(True)
            v,t,f = ControlData._parse_value(default)
            types.append(t)
            formats.append(f)
            cast_defaults.append(v)
        return pandas.DataFrame({"name":names,"type":types,
                                     "value":cast_defaults,"required":required,
                                    "format":formats})


    @staticmethod
    def _parse_value(value):
        try:
            v = int(value)
            t = np.int32
            f = IFMT
        except Exception as e:
            try:
                v = float(value)
                t = np.float64
                f = FFMT
            except Exception as ee:
                v = value.lower()
                t = str
                f = SFMT
        return v,t,f


    def parse_values_from_lines(self,lines,iskeyword=False):
        """ cast the string lines for a pest control file into actual inputs

        Parameters
        ----------
        lines : list
            strings from pest control file

        """

        if iskeyword:
            extra = {}
            for line in lines:
                raw = line.strip().split()
                if len(raw) == 0 or raw[0] == "#":
                    continue
                name = raw[0].strip().lower()

                value = raw[1].strip()
                v,t,f = self._parse_value(value)
                if name not in self._df.index:
                    extra[name] = v
                else:
                    # if the parsed values type isn't right
                    if t != self._df.loc[name, "type"]:

                        # if a float was expected and int return, not a problem
                        if t == np.int32 and self._df.loc[name, "type"] == np.float64:
                            self._df.loc[name, "value"] = np.float64(v)


                        # if this is a required input, throw
                        elif self._df.loc[name, "required"]:
                            raise Exception("wrong type found for variable " + name + ":" + str(t))
                        else:

                            # else, since this problem is usually a string, check for acceptable values
                            found = False
                            for nname, avalues in self.accept_values.items():
                                if v in avalues:
                                    if t == self._df.loc[nname, "type"]:
                                        self._df.loc[nname, "value"] = v
                                        found = True
                                        break
                            if not found:
                                warnings.warn("non-conforming value found for " + \
                                              name + ":" + str(v) + "...ignoring", PyemuWarning)


                    else:
                        self._df.loc[name, "value"] = v
            return extra




        assert len(lines) == len(CONTROL_VARIABLE_LINES),\
        "ControlData error: len of lines not equal to " +\
        str(len(CONTROL_VARIABLE_LINES))

        for iline,line in enumerate(lines):
            vals = line.strip().split()
            names = CONTROL_VARIABLE_LINES[iline].strip().split()
            for name,val in zip(names,vals):
                v,t,f = self._parse_value(val)
                name = name.replace('[','').replace(']','')
                
                #if the parsed values type isn't right
                if t != self._df.loc[name,"type"]:

                    # if a float was expected and int return, not a problem
                    if t == np.int32 and self._df.loc[name,"type"] == np.float64:
                        self._df.loc[name,"value"] = np.float64(v)


                    # if this is a required input, throw
                    elif self._df.loc[name,"required"]:
                        raise Exception("wrong type found for variable " + name + ":" + str(t))
                    else:
                        
                        #else, since this problem is usually a string, check for acceptable values
                        found = False
                        for nname,avalues in self.accept_values.items():
                            if v in avalues:
                                if t == self._df.loc[nname,"type"]:
                                    self._df.loc[nname,"value"] = v
                                    found = True
                                    break
                        if not found:
                            warnings.warn("non-conforming value found for " +\
                                  name + ":" + str(v) + "...ignoring",PyemuWarning)


                else:
                    self._df.loc[name,"value"] = v


    def copy(self):
        cd = ControlData()
        cd._df = self._df
        return cd


    @property
    def formatted_values(self):
        """ list the entries and current values in the control data section

        Returns
        -------
        formatted_values : pandas.Series

        """
        return self._df.apply(lambda x: self.formatters[x["type"]](x["value"]),axis=1)


    def write_keyword(self,f):
        kw = super(ControlData, self).__getattribute__("keyword_accessed")
        f.write("* control data keyword\n")
        for n,v in zip(self._df.name,self.formatted_values):
            if n not in kw:
                continue
            f.write("{0:30} {1}\n".format(n,v))


    def write(self,f):
        """ write control data section to a file
        
        Parameters
        ----------
        f: file handle or string filename

        """
        if isinstance(f,str):
            f = open(f,'w')
            f.write("pcf\n")
            f.write("* control data\n")
        for line in CONTROL_VARIABLE_LINES:
            [f.write(self.formatted_values[name.replace('[','').replace(']','')]) for name in line.split()]
            f.write('\n')





"""pyEMU: python modules for Environmental Model Uncertainty analyses.  These
modules are designed to work directly and seamlessly with PEST and PEST++ model
independent interface.  pyEMU can also be used to setup this interface.

Several forms of uncertainty analyses are support including FOSM-based
analyses (pyemu.Schur and pyemu.ErrVar), Monte Carlo (including
GLUE and null-space Monte Carlo) and a prototype iterative Ensemble Smoother

"""

from .ensemble_method import *
from .smoother import *
from .moouu import *
from .pst_from import *
from .da import *




import os
import shutil
import numpy as np
import pandas as pd
from datetime import datetime
import string
from pyemu.logger import Logger
from pyemu.pst import pst_utils
font = {'size'   : 6}
try:
    import matplotlib
    matplotlib.rc("font",**font)

    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages
    from matplotlib.gridspec import GridSpec
except Exception as e:
    raise Exception("error importing matplotlib: {0}".format(str(e)))

import pyemu
figsize=(8,10.5)
nr,nc = 4,2
#page_gs = GridSpec(nr,nc)

abet = string.ascii_uppercase

def plot_summary_distributions(df,ax=None,label_post=False,label_prior=False,
                               subplots=False,figsize=(11,8.5),pt_color='b'):
    """ helper function to plot gaussian distrbutions from prior and posterior
    means and standard deviations

    Parameters
    ----------
    df : pandas.DataFrame
        a dataframe and csv file.  Must have columns named:
        'prior_mean','prior_stdev','post_mean','post_stdev'.  If loaded
        from a csv file, column 0 is assumed to tbe the index
    ax: matplotlib.pyplot.axis
        If None, and not subplots, then one is created
        and all distributions are plotted on a single plot
    label_post: bool
        flag to add text labels to the peak of the posterior
    label_prior: bool
        flag to add text labels to the peak of the prior
    subplots: (boolean)
        flag to use subplots.  If True, then 6 axes per page
        are used and a single prior and posterior is plotted on each
    figsize: tuple
        matplotlib figure size

    Returns
    -------
    figs : list
        list of figures
    axes : list
        list of axes

    Note
    ----
    This is useful for demystifying FOSM results

    if subplots is False, a single axis is returned

    Example
    -------
    ``>>>import matplotlib.pyplot as plt``

    ``>>>import pyemu``

    ``>>>pyemu.helpers.plot_summary_distributions("pest.par.usum.csv")``

    ``>>>plt.show()``
    """
    import matplotlib.pyplot as plt
    if isinstance(df,str):
        df = pd.read_csv(df,index_col=0)
    if ax is None and not subplots:
        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(111)
        ax.grid()


    if "post_stdev" not in df.columns and "post_var" in df.columns:
        df.loc[:,"post_stdev"] = df.post_var.apply(np.sqrt)
    if "prior_stdev" not in df.columns and "prior_var" in df.columns:
        df.loc[:,"prior_stdev"] = df.prior_var.apply(np.sqrt)
    if "prior_expt" not in df.columns and "prior_mean" in df.columns:
        df.loc[:,"prior_expt"] = df.prior_mean
    if "post_expt" not in df.columns and "post_mean" in df.columns:
        df.loc[:,"post_expt"] = df.post_mean

    if subplots:
        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(2,3,1)
        ax_per_page = 6
        ax_count = 0
        axes = []
        figs = []
    for name in df.index:
        x,y = gaussian_distribution(df.loc[name,"post_expt"],
                                    df.loc[name,"post_stdev"])
        ax.fill_between(x,0,y,facecolor=pt_color,edgecolor="none",alpha=0.25)
        if label_post:
            mx_idx = np.argmax(y)
            xtxt,ytxt = x[mx_idx],y[mx_idx] * 1.001
            ax.text(xtxt,ytxt,name,ha="center",alpha=0.5)

        x,y = gaussian_distribution(df.loc[name,"prior_expt"],
                                    df.loc[name,"prior_stdev"])
        ax.plot(x,y,color='0.5',lw=3.0,dashes=(2,1))
        if label_prior:
            mx_idx = np.argmax(y)
            xtxt,ytxt = x[mx_idx],y[mx_idx] * 1.001
            ax.text(xtxt,ytxt,name,ha="center",alpha=0.5)
        #ylim = list(ax.get_ylim())
        #ylim[1] *= 1.2
        #ylim[0] = 0.0
        #ax.set_ylim(ylim)
        if subplots:
            ax.set_title(name)
            ax_count += 1
            ax.set_yticklabels([])
            axes.append(ax)
            if name == df.index[-1]:
                break
            if ax_count >= ax_per_page:
                figs.append(fig)
                fig = plt.figure(figsize=figsize)
                ax_count = 0
            ax = plt.subplot(2,3,ax_count+1)
    if subplots:
        figs.append(fig)
        return figs, axes
    ylim = list(ax.get_ylim())
    ylim[1] *= 1.2
    ylim[0] = 0.0
    ax.set_ylim(ylim)
    ax.set_yticklabels([])
    return ax

def gaussian_distribution(mean, stdev, num_pts=50):
    """ get an x and y numpy.ndarray that spans the +/- 4
    standard deviation range of a gaussian distribution with
    a given mean and standard deviation. useful for plotting

    Parameters
    ----------
    mean : float
        the mean of the distribution
    stdev : float
        the standard deviation of the distribution
    num_pts : int
        the number of points in the returned ndarrays.
        Default is 50

    Returns
    -------
    x : numpy.ndarray
        the x-values of the distribution
    y : numpy.ndarray
        the y-values of the distribution

    """
    xstart = mean - (4.0 * stdev)
    xend = mean + (4.0 * stdev)
    x = np.linspace(xstart,xend,num_pts)
    y = (1.0/np.sqrt(2.0*np.pi*stdev*stdev)) * np.exp(-1.0 * ((x - mean)**2)/(2.0*stdev*stdev))
    return x,y

def pst_helper(pst,kind=None,**kwargs):

    echo = kwargs.get("echo",False)
    logger = pyemu.Logger("plot_pst_helper.log",echo=echo)
    logger.statement("plot_utils.pst_helper()")

    kinds = {"prior":pst_prior,"1to1":res_1to1,"obs_v_sim":res_obs_v_sim,
             "phi_pie":res_phi_pie,"weight_hist":pst_weight_hist,
             "phi_progress":phi_progress}

    if kind is None:
        returns = []
        base_filename = pst.filename
        if pst.new_filename is not None:
            base_filename = pst.new_filename
        base_filename = base_filename.replace(".pst",'')
        for name,func in kinds.items():
            plt_name = base_filename+"."+name+".pdf"
            returns.append(func(pst,logger=logger,filename=plt_name))

        return returns
    elif kind not in kinds:
        logger.lraise("unrecognized kind:{0}, should one of {1}"
                      .format(kind,','.join(list(kinds.keys()))))
    return kinds[kind](pst, logger, **kwargs)


def phi_progress(pst,logger=None,filename=None,**kwargs):
    """ make plot of phi vs number of model runs - requires
    available pestpp .iobj file
        Parameters
        ----------
        pst : pyemu.Pst
        logger : Logger
            if None, a generic one is created.  Default is None
        filename : str
            PDF filename to save figures to.  If None, figures are returned.  Default is None
        kwargs : dict
            optional keyword args to pass to plotting functions


        """
    if logger is None:
        logger = Logger('Default_Loggger.log', echo=False)
    logger.log("plot phi_progress")

    iobj_file = pst.filename.replace(".pst",".iobj")
    if not os.path.exists(iobj_file):
        logger.lraise("couldn't find iobj file {0}".format(iobj_file))
    df = pd.read_csv(iobj_file)
    if "ax" in kwargs:
        ax = kwargs["ax"]
    else:
        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(1,1,1)
    ax.plot(df.model_runs_completed,df.total_phi,marker='.')
    ax.set_xlabel("model runs")
    ax.set_ylabel("$\phi$")
    ax.grid()
    if filename is not None:
        plt.savefig(filename)
    logger.log("plot phi_progress")
    return ax



def get_page_axes(count=nr*nc):
    axes = [plt.subplot(nr,nc,i+1) for i in range(min(count,nr*nc))]
    #[ax.set_yticks([]) for ax in axes]
    return axes

def res_1to1(pst,logger=None,filename=None,plot_hexbin=False,histogram=False,**kwargs):
    """ make 1-to-1 plots and also observed vs residual by observation group
    Parameters
    ----------
    pst : pyemu.Pst
    logger : Logger
        if None, a generic one is created.  Default is None
    filename : str
        PDF filename to save figures to.  If None, figures are returned.  Default is None
    kwargs : dict
        optional keyword args to pass to plotting functions

    TODO: color symbols by weight


    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot res_1to1")

    if "ensemble" in kwargs:
        try:
            res=pst_utils.res_from_en(pst,kwargs['ensemble'])
        except Exception as e:
            logger.lraise("res_1to1: error loading ensemble file: {0}".format( str(e)))
    else:
        try:
            res = pst.res
        except:
            logger.lraise("res_phi_pie: pst.res is None, couldn't find residuals file")

    obs = pst.observation_data

    if "grouper" in kwargs:
        raise NotImplementedError()
    else:
        grouper = obs.groupby(obs.obgnme).groups

    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5, 0.5, "pyemu.Pst.plot(kind='1to1')\nfrom pest control file '{0}'\n at {1}"
                    .format(pst.filename, str(datetime.now())), ha="center")
    #if plot_hexbin:
    #    pdfname = pst.filename.replace(".pst", ".1to1.hexbin.pdf")
    #else:
    #    pdfname = pst.filename.replace(".pst", ".1to1.pdf")
    figs = []
    ax_count = 0
    for g, names in grouper.items():
        logger.log("plotting 1to1 for {0}".format(g))

        obs_g = obs.loc[names, :]
        obs_g.loc[:, "sim"] = res.loc[names, "modelled"]
        logger.statement("using control file obsvals to calculate residuals")
        obs_g.loc[:,'res'] = obs_g.sim - obs_g.obsval
        if "include_zero" not in kwargs or kwargs["include_zero"] is True:
            obs_g = obs_g.loc[obs_g.weight > 0, :]
        if obs_g.shape[0] == 0:
            logger.statement("no non-zero obs for group '{0}'".format(g))
            logger.log("plotting 1to1 for {0}".format(g))
            continue

        if ax_count % (nr * nc) == 0:
            if ax_count > 0:
                plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig = plt.figure(figsize=figsize)
            axes = get_page_axes()
            ax_count = 0

        ax = axes[ax_count]

        #if obs_g.shape[0] == 1:
        #    ax.scatter(list(obs_g.sim),list(obs_g.obsval),marker='.',s=30,color='b')
        #else:
        mx = max(obs_g.obsval.max(), obs_g.sim.max())
        mn = min(obs_g.obsval.min(), obs_g.sim.min())

        #if obs_g.shape[0] == 1:
        mx *= 1.1
        mn *= 0.9
        ax.axis('square')
        if plot_hexbin:
            ax.hexbin(obs_g.obsval.values, obs_g.sim.values, mincnt=1, gridsize=(75, 75),
                      extent=(mn, mx, mn, mx), bins='log', edgecolors=None)
#               plt.colorbar(ax=ax)
        else:
            ax.scatter([obs_g.obsval], [obs_g.sim], marker='.', s=10, color='b')



        ax.plot([mn,mx],[mn,mx],'k--',lw=1.0)
        xlim = (mn,mx)
        ax.set_xlim(mn,mx)
        ax.set_ylim(mn,mx)
        ax.grid()

        ax.set_xlabel("observed",labelpad=0.1)
        ax.set_ylabel("simulated",labelpad=0.1)
        ax.set_title("{0}) group:{1}, {2} observations".
                                 format(abet[ax_count], g, obs_g.shape[0]), loc="left")

        ax_count += 1

        if histogram==False:
            ax = axes[ax_count]
            ax.scatter(obs_g.obsval, obs_g.res, marker='.', s=10, color='b')
            ylim = ax.get_ylim()
            mx = max(np.abs(ylim[0]), np.abs(ylim[1]))
            if obs_g.shape[0] == 1:
                mx *= 1.1
            ax.set_ylim(-mx, mx)
            #show a zero residuals line
            ax.plot(xlim, [0,0], 'k--', lw=1.0)
            meanres= obs_g.res.mean()
            # show mean residuals line
            ax.plot(xlim,[meanres,meanres], 'r-', lw=1.0)
            ax.set_xlim(xlim)
            ax.set_ylabel("residual",labelpad=0.1)
            ax.set_xlabel("observed",labelpad=0.1)
            ax.set_title("{0}) group:{1}, {2} observations".
                         format(abet[ax_count], g, obs_g.shape[0]), loc="left")
            ax.grid()
            ax_count += 1
        else:
            ax = axes[ax_count]
            ax.hist(obs_g.res, 50, color='b')
            meanres= obs_g.res.mean()
            ax.axvline(meanres, color='r', lw=1)
            b,t = ax.get_ylim()
            ax.text(meanres + meanres/10,
                     t - t/10,
                     'Mean: {:.2f}'.format(meanres))
            ax_count += 1
        logger.log("plotting 1to1 for {0}".format(g))

    for a in range(ax_count, nr * nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])

    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        with PdfPages(filename) as pdf:
            for fig in figs:
                pdf.savefig(fig)
                plt.close(fig)
        logger.log("plot res_1to1")
    else:
        logger.log("plot res_1to1")
        return figs

def res_obs_v_sim(pst,logger=None, filename=None,  **kwargs):
    """
    timeseries plot helper...in progress

    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot res_obs_v_sim")

    if "ensemble" in kwargs:
        try:
            res=pst_utils.res_from_en(pst,kwargs['ensemble'])
        except:
            logger.statement("res_1to1: could not find ensemble file {0}".format(kwargs['ensemble']))
    else:
        try:
            res = pst.res
        except:
            logger.lraise("res_phi_pie: pst.res is None, couldn't find residuals file")

    obs = pst.observation_data

    if "grouper" in kwargs:
        raise NotImplementedError()
    else:
        grouper = obs.groupby(obs.obgnme).groups

    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5, 0.5, "pyemu.Pst.plot(kind='obs_v_sim')\nfrom pest control file '{0}'\n at {1}"
                    .format(pst.filename, str(datetime.now())), ha="center")
    figs = []
    ax_count = 0
    axes = None
    for g, names in grouper.items():
        logger.log("plotting obs_v_sim for {0}".format(g))

        obs_g = obs.loc[names, :]
        obs_g.loc[:, "sim"] = res.loc[names, "modelled"]
        if "include_zero" not in kwargs or kwargs["include_zero"] is False:
            obs_g = obs_g.loc[obs_g.weight > 0, :]

        if obs_g.shape[0] == 0:
            logger.statement("no non-zero obs for group '{0}'".format(g))
            logger.log("plotting obs_v_sim for {0}".format(g))
            continue

        # parse datetimes
        # suffix in decreasing magnitude (yearmonthday)
        try:
            obs_g.loc[:, "datetime_str"] = obs_g.obsnme.apply(lambda x: x.split('_')[-1])
        except:
            logger.statement("res_obs_v_sim error forming datetime_str")
            continue
        # Default datetime
        try:
            obs_g.loc[:, "datetime"] = pd.to_datetime(obs_g.datetime_str,format="%Y%m%d")
        except:
            logger.statement("res_obs_v_sim error casting datetime using default {0}".format(g))

        # abbreviated date format ALWAY 2 digit year, mabye followed by 2 digit month, maybe followed by 2 digit day
        obs_g['datetime']=0
        try:
            obs_g.loc[obs_g['datetime_str'].str.len()==2,'datetime']=pd.to_datetime(obs_g['datetime_str']+'1231',format='%y%m%d')
        except:
            logger.statement("res_obs_v_sim error casting datetime using %y {0}".format(g))
        try:
            obs_g.loc[obs_g['datetime_str'].str.len()==4,'datetime']=pd.to_datetime(obs_g['datetime_str'].astype(str)+'01',format='%y%m%d')+datetime.timedelta(months=1)-datetime.timedelta(days=1)
        except:
            logger.statement("res_obs_v_sim error casting datetime using %y%m {0}".format(g))
        try:
            obs_g.loc[obs_g['datetime_str'].str.len()==6,'datetime']=pd.to_datetime(obs_g['datetime_str'].astype(str),format='%y%m%d')
        except:
            logger.statement("res_obs_v_sim error casting datetime using %y%m%d {0}".format(g))


        if ax_count % (nr * nc) == 0:
            plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig = plt.figure(figsize=figsize)
            axes = get_page_axes()
            ax_count = 0

        ax = axes[ax_count]
        obs_g.loc[:,"site"] = obs_g.obsnme.apply(lambda x: x.split('_')[0])
        for site in obs_g.site.unique():
            obs_s = obs_g.loc[obs_g.site==site,:]
            obs_s.sort_values(by="datetime")
            ax.plot(obs_s.datetime, obs_s.obsval, ls='-', marker='.', ms=10, color='b')
            ax.plot(obs_s.datetime, obs_s.sim, ls='-', marker='.', ms=10, color='0.5')
        ax.set_xlim(obs_g.datetime.min(),obs_g.datetime.max())
        ax.grid()
        ax.set_xlabel("datetime",labelpad=0.1)
        ax.set_title("{0}) group:{1}, {2} observations".
                     format(abet[ax_count], g, names.shape[0]), loc="left")
        ax_count += 1
        logger.log("plotting obs_v_sim for {0}".format(g))

    if axes is None:
        return

    for a in range(ax_count,nr*nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])



    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        with PdfPages(pst.filename.replace(".pst", ".obs_v_sim.pdf")) as pdf:
            for fig in figs:
                pdf.savefig(fig)
                plt.close(fig)
                logger.log("plot res_obs_v_sim")
    else:
        logger.log("plot res_obs_v_sim")
        return figs

def plot_id_bar(id_df, nsv=None, logger=None, **kwargs):
    """
    Plot a stacked bar chart of identifiability based on a identifiability dataframe

    Parameters
    ----------
        id_df : pandas dataframe of identifiability
        nsv : number of singular values to consider
        logger : pyemu.Logger
        kwargs : dict of keyword arguments

    Returns
    -------
    ax : matplotlib.Axis

       Example
    -------
    ``>>> import pyemu``
    ``>>> pest_obj = pyemu.Pst(pest_control_file)``
    ``>>> ev = pyemu.ErrVar(jco='freyberg_jac.jcb'))``
    ``>>> id_df = ev.get_identifiability_dataframe(singular_value=48)``
    ``>>> pyemu.plot_id_bar(id_df, nsv=12, figsize=(12,4)``
     """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot id bar")


    df = id_df.copy()

    # drop the final `ident` column
    if 'ident' in df.columns:
        df.drop('ident', inplace=True, axis=1)

    if nsv is None or nsv > len(df.columns):
        nsv = len(df.columns)
        logger.log('set number of SVs and number in the dataframe')

    df = df[df.columns[:nsv]]

    df['ident'] = df.sum(axis=1)
    df.sort_values(by='ident', inplace=True, ascending=False)
    df.drop('ident', inplace=True, axis=1)

    if 'figsize' in kwargs:
        figsize=kwargs['figsize']
    else:
        figsize = (8, 10.5)
    if "ax" in kwargs:
        ax = kwargs["ax"]
    else:
        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(1,1,1)

    # plto the stacked bar chart (the easy part!)
    df.plot.bar(stacked=True, cmap='jet_r', legend=False, ax=ax)

    #
    # horrible shenanigans to make a colorbar rather than a legend
    #

    # special case colormap just dark red if one SV
    if nsv == 1:
        tcm = matplotlib.colors.LinearSegmentedColormap.from_list('one_sv', [plt.get_cmap('jet_r')(0)] * 2, N=2)
        sm = plt.cm.ScalarMappable(cmap=tcm, norm=matplotlib.colors.Normalize(vmin=0, vmax=nsv + 1))
    # or typically just rock the jet_r colormap over the range of SVs
    else:
        sm = plt.cm.ScalarMappable(cmap=plt.get_cmap('jet_r'), norm=matplotlib.colors.Normalize(vmin=1, vmax=nsv))
    sm._A = []

    # now, if too many ticks for the colorbar, summarize them
    if nsv < 20:
        ticks = range(1, nsv + 1)
    else:
        ticks = np.arange(1, nsv + 1, int((nsv + 1) / 30))

    cb = plt.colorbar(sm)
    cb.set_ticks(ticks)

    logger.log('plot id bar')

    return ax

def res_phi_pie(pst,logger=None, **kwargs):
    """plot current phi components as a pie chart.

    Parameters
    ----------
    pst : pyemu.Pst
    logger : pyemu.Logger
    kwargs : dict
        accepts 'include_zero' as a flag to include phi groups with
        only zero-weight obs (not sure why anyone would do this, but
        whatevs).
    Returns
    -------
    ax : matplotlib.Axis


    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot res_phi_pie")

    if "ensemble" in kwargs:
        try:
            res=pst_utils.res_from_en(pst,kwargs['ensemble'])
        except:
            logger.statement("res_1to1: could not find ensemble file {0}".format(kwargs['ensemble']))
    else:
        try:
            res = pst.res
        except:
            logger.lraise("res_phi_pie: pst.res is None, couldn't find residuals file")

    obs = pst.observation_data
    phi = pst.phi
    phi_comps = pst.phi_components
    norm_phi_comps = pst.phi_components_normalized
    keys = list(phi_comps.keys())
    if "include_zero" not in kwargs or kwargs["include_zero"] is True:
        phi_comps = {k:phi_comps[k] for k in keys if phi_comps[k] > 0.0}
        keys = list(phi_comps.keys())
        norm_phi_comps = {k:norm_phi_comps[k] for k in keys}
    if "ax" in kwargs:
        ax = kwargs["ax"]
    else:
        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(1,1,1,aspect="equal")
    labels = ["{0}\n{1:4G}\n({2:3.1f}%)".format(k,phi_comps[k],100. * (phi_comps[k] / phi)) for k in keys]
    ax.pie([float(norm_phi_comps[k]) for k in keys],labels=labels)
    logger.log("plot res_phi_pie")
    if "filename" in kwargs:
        plt.savefig(kwargs["filename"])
    return ax



def pst_weight_hist(pst,logger, **kwargs):
    #raise NotImplementedError()
    pass



def pst_prior(pst,logger=None, filename=None, **kwargs):
    """ helper to plot prior parameter histograms implied by
    parameter bounds. Saves a multipage pdf named <case>.prior.pdf

    Parameters
    ----------
    pst : pyemu.Pst
    logger : pyemu.Logger
    filename : str
        PDF filename to save plots to. If None, return figs without saving.  Default is None.
    kwargs : dict
        accepts 'grouper' as dict to group parameters on to a single axis (use
        parameter groups if not passed),
        'unqiue_only' to only show unique mean-stdev combinations within a
        given group

    Returns
    -------
    None

    TODO
    ----
    external parcov, unique mean-std pairs

    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot pst_prior")
    par = pst.parameter_data

    if "parcov_filename" in pst.pestpp_options:
        logger.warn("ignoring parcov_filename, using parameter bounds for prior cov")
    logger.log("loading cov from parameter data")
    cov = pyemu.Cov.from_parameter_data(pst)
    logger.log("loading cov from parameter data")

    logger.log("building mean parameter values")
    li = par.partrans.loc[cov.names] == "log"
    mean = par.parval1.loc[cov.names]
    info = par.loc[cov.names,:].copy()
    info.loc[:,"mean"] = mean
    info.loc[li,"mean"] = mean[li].apply(np.log10)
    logger.log("building mean parameter values")

    logger.log("building stdev parameter values")
    if cov.isdiagonal:
        std = cov.x.flatten()
    else:
        std = np.diag(cov.x)
    std = np.sqrt(std)
    info.loc[:,"prior_std"] = std

    logger.log("building stdev parameter values")

    if std.shape != mean.shape:
        logger.lraise("mean.shape {0} != std.shape {1}".
                      format(mean.shape,std.shape))

    if "grouper" in kwargs:
        raise NotImplementedError()
        #check for consistency here

    else:
        par_adj = par.loc[par.partrans.apply(lambda x: x in ["log","none"]),:]
        grouper = par_adj.groupby(par_adj.pargp).groups
        #grouper = par.groupby(par.pargp).groups

    if len(grouper) == 0:
        raise Exception("no adustable parameters to plot")

    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5,0.5,"pyemu.Pst.plot(kind='prior')\nfrom pest control file '{0}'\n at {1}"
                 .format(pst.filename,str(datetime.now())),ha="center")
    figs = []
    ax_count = 0
    grps_names = list(grouper.keys())
    grps_names.sort()
    for g in grps_names:
        names = grouper[g]
        logger.log("plotting priors for {0}".
                   format(','.join(list(names))))
        if ax_count % (nr * nc) == 0:
            plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig  = plt.figure(figsize=figsize)
            axes = get_page_axes()
            ax_count = 0

        islog = False
        vc = info.partrans.value_counts()
        if vc.shape[0] > 1:
            logger.warn("mixed partrans for group {0}".format(g))
        elif "log" in vc.index:
            islog = True
        ax = axes[ax_count]
        if "unique_only" in kwargs and kwargs["unique_only"]:


            ms = info.loc[names,:].apply(lambda x: (x["mean"],x["prior_std"]),axis=1).unique()
            for (m,s) in ms:
                x, y = gaussian_distribution(m, s)
                ax.fill_between(x, 0, y, facecolor='0.5', alpha=0.5,
                                edgecolor="none")


        else:
            for m,s in zip(info.loc[names,'mean'],info.loc[names,'prior_std']):
                x,y = gaussian_distribution(m,s)
                ax.fill_between(x,0,y,facecolor='0.5',alpha=0.5,
                                            edgecolor="none")
        ax.set_title("{0}) group:{1}, {2} parameters".
                                 format(abet[ax_count],g,names.shape[0]),loc="left")

        ax.set_yticks([])
        if islog:
            ax.set_xlabel("$log_{10}$ parameter value",labelpad=0.1)
        else:
            ax.set_xlabel("parameter value", labelpad=0.1)
        logger.log("plotting priors for {0}".
                   format(','.join(list(names))))

        ax_count += 1

    for a in range(ax_count,nr*nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])

    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        with PdfPages(filename) as pdf:
            plt.tight_layout()
            pdf.savefig(fig)
            plt.close(fig)
        logger.log("plot pst_prior")
    else:
        logger.log("plot pst_prior")
        return figs


def ensemble_helper(ensemble,bins=10,facecolor='0.5',plot_cols=None,
                    filename=None,func_dict = None,
                    sync_bins=True,deter_vals=None,std_window=None,
                    deter_range=False,**kwargs):
    """helper function to plot ensemble histograms

    Parameters
    ----------
    ensemble : varies
        the ensemble argument can be a pandas.DataFrame or derived type or a str, which
        is treated as a filename.  Optionally, ensemble can be a list of these types or
        a dict, in which case, the keys are treated as facecolor str (e.g., 'b', 'y', etc).
    facecolor : str
        the histogram facecolor.  Only applies if ensemble is a single thing
    plot_cols : enumerable
        a collection of columns (in form of a list of parameters, or a dict with keys for
        parsing plot axes and values of parameters) from the ensemble(s) to plot.  If None,
        (the union of) all cols are plotted. Default is None
    filename : str
        the name of the pdf to create.  If None, return figs without saving.  Default is None.
    func_dict : dict
        a dictionary of unary functions (e.g., np.log10_ to apply to columns.  Key is
        column name.  Default is None
    sync_bins : bool
        flag to use the same bin edges for all ensembles. Only applies if more than
        one ensemble is being plotted.  Default is True
    deter_vals : dict
        dict of deterministic values to plot as a vertical line. key is ensemble columnn name
    std_window : float
        the number of standard deviations around the mean to mark as vertical lines.  If None,
        nothing happens.  Default is None
    deter_range : bool
        flag to set xlims to deterministic value +/- std window.  If True, std_window must not be None.
        Default is False

    """
    logger = pyemu.Logger("ensemble_helper.log")
    logger.log("pyemu.plot_utils.ensemble_helper()")
    ensembles = _process_ensemble_arg(ensemble,facecolor,logger)

    #apply any functions
    if func_dict is not None:
        logger.log("applying functions")
        for col,func in func_dict.items():
            for fc,en in ensembles.items():
                if col in en.columns:
                    en.loc[:,col] = en.loc[:,col].apply(func)
        logger.log("applying functions")


    #get a list of all cols (union)
    all_cols = set()
    for fc, en in ensembles.items():
        cols = set(en.columns)
        all_cols.update(cols)
    if plot_cols is None:
        plot_cols = {i: [v] for i, v in (zip(all_cols, all_cols))}
    else:
        if isinstance(plot_cols,list):
            splot_cols = set(plot_cols)
            plot_cols = {i: [v] for i, v in (zip(plot_cols, plot_cols))}
        elif isinstance(plot_cols,dict):
            splot_cols = []
            for label,pcols in plot_cols.items():
                splot_cols.extend(list(pcols))
            splot_cols = set(splot_cols)
        else:
            logger.lraise("unrecognized plot_cols type: {0}, should be list or dict".
                          format(type(plot_cols)))

        missing = splot_cols - all_cols
        if len(missing) > 0:
            logger.lraise("the following plot_cols are missing: {0}".
                          format(','.join(missing)))

    logger.statement("plotting {0} histograms".format(len(plot_cols)))

    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5, 0.5, "pyemu.plot_utils.ensemble_helper()\n at {0}"
                    .format(str(datetime.now())), ha="center")
    #plot_cols = list(plot_cols)
    #plot_cols.sort()
    labels = list(plot_cols.keys())
    labels.sort()
    logger.statement("saving pdf to {0}".format(filename))
    figs = []

    ax_count = 0

    #for label,plot_col in plot_cols.items():
    for label in labels:
        plot_col = plot_cols[label]
        logger.log("plotting reals for {0}".format(label))
        if ax_count % (nr * nc) == 0:
            plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig = plt.figure(figsize=figsize)
            axes = get_page_axes()
            [ax.set_yticks([]) for ax in axes]
            ax_count = 0

        ax = axes[ax_count]
        ax.set_title("{0}) {1}".format(abet[ax_count],label),loc="left")
        if sync_bins:
            mx,mn = -1.0e+30,1.0e+30
            for fc,en in ensembles.items():
                # for pc in plot_col:
                #     if pc in en.columns:
                #         emx,emn = en.loc[:,pc].max(),en.loc[:,pc].min()
                #         mx = max(mx,emx)
                #         mn = min(mn,emn)
                emn = en.loc[:,plot_col].values.min()
                emx = en.loc[:, plot_col].values.max()
                mx = max(mx, emx)
                mn = min(mn,emn)
            plot_bins = np.linspace(mn,mx,num=bins)
            logger.statement("{0} min:{1:5G}, max:{2:5G}".format(label,mn,mx))
        else:
            plot_bins=bins
        for fc,en in ensembles.items():
            #for pc in plot_col:
            #    if pc in en.columns:
            #        try:
            #            en.loc[:,pc].hist(bins=plot_bins,facecolor=fc,
            #                                    edgecolor="none",alpha=0.5,
            #                                    normed=True,ax=ax)
            #        except Exception as e:
            #            logger.warn("error plotting histogram for {0}:{1}".
            #                        format(pc,str(e)))
            vals = en.loc[:,plot_col].values.flatten()
            #print(plot_bins)
            #print(vals)
            ax.hist(vals,bins=plot_bins,edgecolor="none",alpha=0.5,normed=True,facecolor=fc)
            v = None
            if deter_vals is not None:
                for pc in plot_col:
                    if pc in deter_vals:
                        ylim = ax.get_ylim()
                        v = deter_vals[pc]
                        ax.plot([v,v],ylim,"k--",lw=1.5)
                        ax.set_ylim(ylim)


            if std_window is not None:
                try:
                    ylim = ax.get_ylim()
                    mn, st = en.loc[:,pc].mean(), en.loc[:,pc].std() * (std_window / 2.0)

                    ax.plot([mn - st, mn - st], ylim, color=fc, lw=1.5,ls='--')
                    ax.plot([mn + st, mn + st], ylim, color=fc, lw=1.5,ls='--')
                    ax.set_ylim(ylim)
                    if deter_range and v is not None:
                        xmn = v - st
                        xmx = v + st
                        ax.set_xlim(xmn,xmx)
                except:
                    logger.warn("error plotting std window for {0}".
                                format(pc))
        ax.grid()

        ax_count += 1

    for a in range(ax_count, nr * nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])

    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        plt.tight_layout()
        with PdfPages(filename) as pdf:
            for fig in figs:
                pdf.savefig(fig)
                plt.close(fig)
    logger.log("pyemu.plot_utils.ensemble_helper()")

def ensemble_change_summary(ensemble1, ensemble2, pst,bins=10, facecolor='0.5',logger=None,filename=None,**kwargs):
    """helper function to plot first and second moment change histograms

    Parameters
    ----------
    ensemble1 : varies
        str or pd.DataFrames
    ensemble2 : varies
        str or pd.DataFrame
    pst : pyemu.Pst
        pst instance
    facecolor : str
        the histogram facecolor.
    filename : str
        the name of the pdf to create. If None, return figs without saving.  Default is None.

    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot ensemble change")

    if isinstance(ensemble1, str):
        ensemble1 = pd.read_csv(ensemble1,index_col=0)
    ensemble1.columns = ensemble1.columns.str.lower()

    if isinstance(ensemble2, str):
        ensemble2 = pd.read_csv(ensemble2,index_col=0)
    ensemble2.columns = ensemble2.columns.str.lower()

    # better to ensure this is caught by pestpp-ies ensemble csvs
    unnamed1 = [col for col in ensemble1.columns if "unnamed:" in col]
    if len(unnamed1) != 0:
        ensemble1 = ensemble1.iloc[:,:-1] # ensure unnamed col result of poor csv read only (ie last col)
    unnamed2 = [col for col in ensemble2.columns if "unnamed:" in col]
    if len(unnamed2) != 0:
        ensemble2 = ensemble2.iloc[:,:-1] # ensure unnamed col result of poor csv read only (ie last col)

    d = set(ensemble1.columns).symmetric_difference(set(ensemble2. columns))

    if len(d) != 0:
        logger.lraise("ensemble1 does not have the same columns as ensemble2: {0}".
                      format(','.join(d)))
    if "grouper" in kwargs:
        raise NotImplementedError()
    else:
        en_cols = set(ensemble1.columns)
        if len(en_cols.symmetric_difference(set(pst.par_names))) == 0:
            par = pst.parameter_data.loc[pst.adj_par_names,:]
            grouper = par.groupby(par.pargp).groups
            grouper["all"] = pst.adj_par_names
            li = par.loc[par.partrans == "log","parnme"]
            ensemble1.loc[:,li] = ensemble1.loc[:,li].apply(np.log10)
            ensemble2.loc[:, li] = ensemble2.loc[:, li].apply(np.log10)
        elif len(en_cols.symmetric_difference(set(pst.obs_names))) == 0:
            obs = pst.observation_data.loc[pst.nnz_obs_names,:]
            grouper = obs.groupby(obs.obgnme).groups
            grouper["all"] = pst.nnz_obs_names
        else:
            logger.lraise("could not match ensemble cols with par or obs...")

    en1_mn, en1_std = ensemble1.mean(axis=0), ensemble1.std(axis=0)
    en2_mn, en2_std = ensemble2.mean(axis=0), ensemble2.std(axis=0)

    # mn_diff = 100.0 * ((en1_mn - en2_mn) / en1_mn)
    # std_diff = 100 * ((en1_std - en2_std) / en1_std)

    mn_diff = -1 * (en2_mn - en1_mn)
    std_diff = 100 * (((en1_std - en2_std) / en1_std))
    #set en1_std==0 to nan
    std_diff[en1_std.index[en1_std==0]] = np.nan



    #diff = ensemble1 - ensemble2
    #mn_diff = diff.mean(axis=0)
    #std_diff = diff.std(axis=0)


    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5, 0.5, "pyemu.Pst.plot(kind='1to1')\nfrom pest control file '{0}'\n at {1}"
                    .format(pst.filename, str(datetime.now())), ha="center")
    #if plot_hexbin:
    #    pdfname = pst.filename.replace(".pst", ".1to1.hexbin.pdf")
    #else:
    #    pdfname = pst.filename.replace(".pst", ".1to1.pdf")
    figs = []
    ax_count = 0
    for g, names in grouper.items():
        logger.log("plotting change for {0}".format(g))

        mn_g = mn_diff.loc[names]
        std_g = std_diff.loc[names]

        if mn_g.shape[0] == 0:
            logger.statement("no entries for group '{0}'".format(g))
            logger.log("plotting change for {0}".format(g))
            continue

        if ax_count % (nr * nc) == 0:
            if ax_count > 0:
                plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig = plt.figure(figsize=figsize)
            axes = get_page_axes()
            ax_count = 0

        ax = axes[ax_count]
        mn_g.hist(ax=ax,facecolor=facecolor,alpha=0.5,edgecolor=None,bins=bins)
        #mx = max(mn_g.max(), mn_g.min(),np.abs(mn_g.max()),np.abs(mn_g.min())) * 1.2
        #ax.set_xlim(-mx,mx)

        #std_g.hist(ax=ax,facecolor='b',alpha=0.5,edgecolor=None)



        #ax.set_xlim(xlim)
        ax.set_yticklabels([])
        ax.set_xlabel("mean change",labelpad=0.1)
        ax.set_title("{0}) mean change group:{1}, {2} entries\nmax:{3:10G}, min:{4:10G}".
                     format(abet[ax_count], g, mn_g.shape[0],mn_g.max(),mn_g.min()), loc="left")
        ax.grid()
        ax_count += 1

        ax = axes[ax_count]
        std_g.hist(ax=ax, facecolor=facecolor, alpha=0.5, edgecolor=None, bins=bins)
        # std_g.hist(ax=ax,facecolor='b',alpha=0.5,edgecolor=None)



        # ax.set_xlim(xlim)
        ax.set_yticklabels([])
        ax.set_xlabel("sigma percent reduction", labelpad=0.1)
        ax.set_title("{0}) sigma change group:{1}, {2} entries\nmax:{3:10G}, min:{4:10G}".
                     format(abet[ax_count], g, mn_g.shape[0], std_g.max(), std_g.min()), loc="left")
        ax.grid()
        ax_count += 1

        logger.log("plotting change for {0}".format(g))

    for a in range(ax_count, nr * nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])

    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        plt.tight_layout()
        with PdfPages(filename) as pdf:
            for fig in figs:
                pdf.savefig(fig)
                plt.close(fig)
        logger.log("plot ensemble change")
    else:
        logger.log("plot ensemble change")
        return figs

def _process_ensemble_arg(ensemble,facecolor, logger):
    ensembles = {}
    if isinstance(ensemble, pd.DataFrame):
        if not isinstance(facecolor, str):
            logger.lraise("facecolor must be str")
        ensembles[facecolor] = ensemble

    elif isinstance(ensemble, str):
        if not isinstance(facecolor, str):
            logger.lraise("facecolor must be str")

        logger.log('loading ensemble from csv file {0}'.format(ensemble))
        en = pd.read_csv(ensemble, index_col=0)
        logger.statement("{0} shape: {1}".format(ensemble, en.shape))
        ensembles[facecolor] = en
        logger.log('loading ensemble from csv file {0}'.format(ensemble))

    elif isinstance(ensemble, list):
        if isinstance(facecolor, list):
            if len(ensemble) != len(facecolor):
                logger.lraise("facecolor list len != ensemble list len")
        else:
            colors = ['m', 'c', 'b', 'r', 'g', 'y']

            facecolor = [colors[i] for i in range(len(ensemble))]
        ensembles = {}
        for fc, en_arg in zip(facecolor, ensemble):
            if isinstance(en_arg, str):
                logger.log("loading ensemble from csv file {0}".format(en_arg))
                en = pd.read_csv(en_arg, index_col=0)
                logger.log("loading ensemble from csv file {0}".format(en_arg))
                logger.statement("ensemble {0} gets facecolor {1}".format(en_arg, fc))

            elif isinstance(en_arg, pd.DataFrame):
                en = en_arg
            else:
                logger.lraise("unrecognized ensemble list arg:{0}".format(en_file))
            ensembles[fc] = en

    elif isinstance(ensemble, dict):
        for fc, en_arg in ensemble.items():
            if isinstance(en_arg, pd.DataFrame):
                ensembles[fc] = en_arg
            elif isinstance(en_arg, str):
                logger.log("loading ensemble from csv file {0}".format(en_arg))
                en = pd.read_csv(en_arg, index_col=0)
                logger.log("loading ensemble from csv file {0}".format(en_arg))
                ensembles[fc] = en
            else:
                logger.lraise("unrecognized ensemble list arg:{0}".format(en_arg))
    try:
        for fc in ensembles:
            ensembles[fc].columns = ensembles[fc].columns.str.lower()
    except:
        logger.lraise("error processing ensemble")

    return ensembles

def ensemble_res_1to1(ensemble, pst,facecolor='0.5',logger=None,filename=None,
                      skip_groups=[],base_ensemble=None,**kwargs):
    """helper function to plot ensemble 1-to-1 plots sbowing the simulated range

    Parameters
    ----------
    ensemble : varies
        the ensemble argument can be a pandas.DataFrame or derived type or a str, which
        is treated as a fileanme.  Optionally, ensemble can be a list of these types or
        a dict, in which case, the keys are treated as facecolor str (e.g., 'b', 'y', etc).
    pst : pyemu.Pst
        pst instance
    facecolor : str
        the histogram facecolor.  Only applies if ensemble is a single thing
    filename : str
        the name of the pdf to create. If None, return figs without saving.  Default is None.
    base_ensemble : varies
        an optional ensemble argument for the observations + noise ensemble.
        This will be plotted as a transparent red bar on the 1to1 plot.

    """
    if logger is None:
        logger=Logger('Default_Loggger.log',echo=False)
    logger.log("plot res_1to1")
    obs = pst.observation_data
    ensembles = _process_ensemble_arg(ensemble,facecolor,logger)

    if base_ensemble is not None:
        base_ensemble = _process_ensemble_arg(base_ensemble,"r",logger)

    if "grouper" in kwargs:
        raise NotImplementedError()
    else:
        grouper = obs.groupby(obs.obgnme).groups
        for skip_group in skip_groups:
            grouper.pop(skip_group)

    fig = plt.figure(figsize=figsize)
    if "fig_title" in kwargs:
        plt.figtext(0.5,0.5,kwargs["fig_title"])
    else:
        plt.figtext(0.5, 0.5, "pyemu.Pst.plot(kind='1to1')\nfrom pest control file '{0}'\n at {1}"
                    .format(pst.filename, str(datetime.now())), ha="center")
    #if plot_hexbin:
    #    pdfname = pst.filename.replace(".pst", ".1to1.hexbin.pdf")
    #else:
    #    pdfname = pst.filename.replace(".pst", ".1to1.pdf")
    figs = []
    ax_count = 0
    for g, names in grouper.items():
        logger.log("plotting 1to1 for {0}".format(g))

        obs_g = obs.loc[names, :]
        logger.statement("using control file obsvals to calculate residuals")
        if "include_zero" not in kwargs or kwargs["include_zero"] is False:
            obs_g = obs_g.loc[obs_g.weight > 0, :]
        if obs_g.shape[0] == 0:
            logger.statement("no non-zero obs for group '{0}'".format(g))
            logger.log("plotting 1to1 for {0}".format(g))
            continue

        if ax_count % (nr * nc) == 0:
            if ax_count > 0:
                plt.tight_layout()
            #pdf.savefig()
            #plt.close(fig)
            figs.append(fig)
            fig = plt.figure(figsize=figsize)
            axes = get_page_axes()
            ax_count = 0

        ax = axes[ax_count]
        if base_ensemble is None:
            mx = obs_g.obsval.max()
            mn =  obs_g.obsval.min()
        else:
            mn = base_ensemble["r"].loc[:,names].min().min()
            mx = base_ensemble["r"].loc[:, names].max().max()
        #if obs_g.shape[0] == 1:
        mx *= 1.1
        mn *= 0.9
        #ax.axis('square')
        if base_ensemble is not None:
            obs_gg = obs_g.sort_values(by="obsval")

            for c, en in base_ensemble.items():
                en_g = en.loc[:, obs_gg.obsnme]
                ex = en_g.max()
                en = en_g.min()
                #[ax.plot([ov, ov], [een, eex], color=c,alpha=0.3) for ov, een, eex in zip(obs_g.obsval.values, en.values, ex.values)]
                ax.fill_between(obs_gg.obsval,en,ex,facecolor=c,alpha=0.2)
        #ax.scatter([obs_g.sim], [obs_g.obsval], marker='.', s=10, color='b')
        for c,en in ensembles.items():
            en_g = en.loc[:,obs_g.obsnme]
            ex = en_g.max()
            en = en_g.min()
            [ax.plot([ov,ov],[een,eex],color=c) for ov,een,eex in zip(obs_g.obsval.values,en.values,ex.values)]


        ax.plot([mn,mx],[mn,mx],'k--',lw=1.0)
        xlim = (mn,mx)
        ax.set_xlim(mn,mx)
        ax.set_ylim(mn,mx)
        if mx > 1.0e5:
            ax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%1.0e'))
            ax.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%1.0e'))
        ax.grid()

        ax.set_xlabel("observed",labelpad=0.1)
        ax.set_ylabel("simulated",labelpad=0.1)
        ax.set_title("{0}) group:{1}, {2} observations".
                                 format(abet[ax_count], g, obs_g.shape[0]), loc="left")

        ax_count += 1
        ax = axes[ax_count]
        #ax.scatter(obs_g.obsval, obs_g.res, marker='.', s=10, color='b')

        if base_ensemble is not None:
            obs_gg = obs_g.sort_values(by="obsval")

            for c, en in base_ensemble.items():
                en_g = en.loc[:, obs_gg.obsnme].subtract(obs_gg.obsval)
                ex = en_g.max()
                en = en_g.min()
                #[ax.plot([ov, ov], [een, eex], color=c,alpha=0.3) for ov, een, eex in zip(obs_g.obsval.values, en.values, ex.values)]
                ax.fill_between(obs_gg.obsval,en,ex,facecolor=c,alpha=0.2)

        for c,en in ensembles.items():
            en_g = en.loc[:,obs_g.obsnme].subtract(obs_g.obsval,axis=1)
            ex = en_g.max()
            en = en_g.min()
            [ax.plot([ov,ov],[een,eex],color=c) for ov,een,eex in zip(obs_g.obsval.values,en.values,ex.values)]
        # if base_ensemble is not None:
        #     if base_ensemble is not None:
        #         for c, en in base_ensemble.items():
        #             en_g = en.loc[:, obs_g.obsnme].subtract(obs_g.obsval,axis=1)
        #             ex = en_g.max()
        #             en = en_g.min()
        #             [ax.plot([ov, ov], [een, eex], color=c, alpha=0.3) for ov, een, eex in
        #              zip(obs_g.obsval.values, en.values, ex.values)]
        ylim = ax.get_ylim()
        mx = max(np.abs(ylim[0]), np.abs(ylim[1]))
        if obs_g.shape[0] == 1:
            mx *= 1.1
        ax.set_ylim(-mx, mx)
        #show a zero residuals line
        ax.plot(xlim, [0,0], 'k--', lw=1.0)

        ax.set_xlim(xlim)
        ax.set_ylabel("residual",labelpad=0.1)
        ax.set_xlabel("observed",labelpad=0.1)
        ax.set_title("{0}) group:{1}, {2} observations".
                     format(abet[ax_count], g, obs_g.shape[0]), loc="left")
        ax.grid()
        if ax.get_xlim()[1] > 1.0e5:
            ax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%1.0e'))

        ax_count += 1

        logger.log("plotting 1to1 for {0}".format(g))

    for a in range(ax_count, nr * nc):
        axes[a].set_axis_off()
        axes[a].set_yticks([])
        axes[a].set_xticks([])

    plt.tight_layout()
    #pdf.savefig()
    #plt.close(fig)
    figs.append(fig)
    if filename is not None:
        plt.tight_layout()
        with PdfPages(filename) as pdf:
            for fig in figs:
                pdf.savefig(fig)
                plt.close(fig)
        logger.log("plot res_1to1")
    else:
        logger.log("plot res_1to1")
        return figs




# def par_cov_helper(cov,pst,logger=None,filename=None,**kwargs):
#     assert isinstance(cov,pyemu.Cov)
#     if logger is None:
#         logger=Logger('Default_Loggger.log',echo=False)
#     logger.log("plot par cov")
#
#     par = pst.parameter_data
#     if "grouper" in kwargs:
#         raise NotImplementedError()
#     else:
#         grouper = par.groupby(par.pargp).groups
#
#     fig = plt.figure(figsize=figsize)
#     if "fig_title" in kwargs:
#         plt.figtext(0.5,0.5,kwargs["fig_title"])
#     else:
#         plt.figtext(0.5, 0.5, "pyemu.Pst.plot(kind='1to1')\nfrom pest control file '{0}'\n at {1}"
#                     .format(pst.filename, str(datetime.now())), ha="center")
#
#     figs = []
#     ax_count = 0
#     for g, names in grouper.items():
#         logger.log("plotting par cov for {0}".format(g))
#         if ax_count % (nr * nc) == 0:
#             if ax_count > 0:
#                 plt.tight_layout()
#             #pdf.savefig()
#             #plt.close(fig)
#             figs.append(fig)
#             fig = plt.figure(figsize=figsize)
#             axes = get_page_axes()
#             ax_count = 0
#
#         ax = axes[ax_count]
#         names = list(names)
#         cov_g = cov.get(row_names=names,col_names=names).x
#         ax.imshow(np.ma.masked_where(cov_g==0,cov_g))
#         ax.set_title("{0}) group:{1}, {2} elements".
#                      format(abet[ax_count], g, cov_g.shape[0]), loc="left")
#
#     plt.tight_layout()
#     # pdf.savefig()
#     # plt.close(fig)
#     figs.append(fig)
#     if filename is not None:
#         plt.tight_layout()
#         with PdfPages(filename) as pdf:
#             for fig in figs:
#                 pdf.savefig(fig)
#                 plt.close(fig)
#         logger.log("plot res_1to1")
#     else:
#         logger.log("plot res_1to1")
#         return figs

def plot_jac_test(csvin, csvout, targetobs=None, filetype=None, maxoutputpages=1, outputdirectory=None):
    """ helper function to plot results of the Jacobian test performed using the pest++ program sweep.

    Parameters
    ----------
    csvin:  string
            name of csv file used as input to sweep, typically developed with
            static method pyemu.helpers.build_jac_test_csv()
    csvout: string
            name of csv file with output generated by sweep, both input
            and output files can be specified in the pest++ control file
            with pyemu using:
            pest_object.pestpp_options["sweep_parameter_csv_file"] = jactest_in_file.csv
            pest_object.pestpp_options["sweep_output_csv_file"] = jactest_out_file.csv
    targetobs: list of strings
            list of observation file names to plot, each parameter used for jactest can
            have up to 32 observations plotted per page, throws a warning if more tha
            10 pages of output are requested per parameter. If none, all observations in
            the output csv file are used.
    filetype: string
            file type to store output, if None, plt.show() is called.
    maxoutputpages: int
            maximum number of pages of output per parameter.  Each page can
            hold up to 32 observation derivatives.  If value > 10, set it to
            10 and throw a warning.  If observations in targetobs > 32*maxoutputpages,
            then a random set is selected from the targetobs list (or all observations
            in the csv file if targetobs=None).
    outputdirectory: path, string, or None
            directory to store results, if None, current working directory is used.
            If string is passed, it is joined to the current working directory and
            created if needed. If os.path is passed, it is used directly.

    Returns
    -------
    None

    Note
    ----
    Used in conjunction with pyemu.helpers.build_jac_test_csv() and sweep to perform
    a Jacobian Test and then view the results. Can generate a lot of plots so easiest
    to put into a separate directory and view the files.

    Example
    -------
    ``>>> import pyemu``
    ``>>> pest_obj = pyemu.Pst(pest_control_file)``
    ``>>> jactest_df = pyemu.helpers.build_jac_test_csv(pst=pest_obj, num_steps=5)``
    ``>>> pest_obj.pestpp_options["sweep_parameter_csv_file"] = "jactest_in.csv"``
    ``>>> pest_obj.pestpp_options["sweep_output_csv_file"] = "jactest_out.csv"``
    ``>>> jactest_df.to_csv(os.path.join(home, "SWEEP", "jactest_in.csv")) ``
    ``>>> pest_obj.write("jactest.pst")
    ``>>> pyemu.helpers.run("{0} {1}".format(sweep_exe, "jactest.pst"), cwd='.', verbose=True)``
    ``>>> pyemu.plot.plot_jac_test("jactest_in.csv", "jactest_out.csv")
    """

    localhome = os.getcwd()
    # check if the output directory exists, if not make it
    if outputdirectory is not None and not os.path.exists(os.path.join(localhome, outputdirectory)):
        os.mkdir(os.path.join(localhome, outputdirectory))
    if outputdirectory is None:
        figures_dir = localhome
    else:
        figures_dir = os.path.join(localhome, outputdirectory)

    # read the input and output files into pandas dataframes
    jactest_in_df = pd.read_csv(csvin, engine='python', index_col=0)
    jactest_in_df.index.name = 'input_run_id'
    jactest_out_df = pd.read_csv(csvout, engine='python', index_col=1)

    # subtract the base run from every row, leaves the one parameter that
    # was perturbed in any row as only non-zero value. Set zeros to nan
    # so round-off doesn't get us and sum across rows to get a column of
    # the perturbation for each row, finally extract to a series. First
    # the input csv and then the output.
    base_par = jactest_in_df.loc['base']
    delta_par_df = jactest_in_df.subtract(base_par, axis='columns')
    delta_par_df.replace(0, np.nan, inplace=True)
    delta_par_df.drop('base', axis='index', inplace=True)
    delta_par_df['change'] = delta_par_df.sum(axis='columns')
    delta_par = pd.Series(delta_par_df['change'])

    base_obs = jactest_out_df.loc['base']
    delta_obs = jactest_out_df.subtract(base_obs)
    delta_obs.drop('base', axis='index', inplace=True)
    # if targetobs is None, then reset it to all the observations.
    if targetobs is None:
        targetobs = jactest_out_df.columns.tolist()[8:]
    delta_obs = delta_obs[targetobs]

    # get the Jacobian by dividing the change in observation by the change in parameter
    # for the perturbed parameters
    jacobian = delta_obs.divide(delta_par, axis='index')

    # use the index created by build_jac_test_csv to get a column of parameter names
    # and increments, then we can plot derivative vs. increment for each parameter
    extr_df = pd.Series(jacobian.index.values).str.extract("(.+)(_\d+$)", expand=True)
    extr_df[1] = pd.to_numeric(extr_df[1].str.replace('_', '')) + 1
    extr_df.rename(columns={0: 'parameter', 1: 'increment'}, inplace=True)
    extr_df.index = jacobian.index

    # make a dataframe for plotting the Jacobian by combining the parameter name
    # and increments frame with the Jacobian frame
    plotframe = pd.concat([extr_df, jacobian], axis=1, join='inner')

    # get a list of observations to keep based on maxoutputpages.
    if maxoutputpages > 10:
        print("WARNING, more than 10 pages of output requested per parameter")
        print("maxoutputpage reset to 10.")
        maxoutputpages=10
    num_obs_plotted = np.min(np.array([maxoutputpages*32, len(targetobs)]))
    if num_obs_plotted < len(targetobs):
        # get random sample
        index_plotted = np.random.choice(len(targetobs), num_obs_plotted, replace=False)
        obs_plotted = [targetobs[x] for x in index_plotted]
        real_pages = maxoutputpages
    else:
        obs_plotted = targetobs
        real_pages = int(targetobs/32) + 1

    # make a subplot of derivative vs. increment one plot for each of the
    # observations in targetobs, and outputs grouped by parameter.
    for param, group in plotframe.groupby('parameter'):
        for page in range(0, real_pages):
            fig, axes = plt.subplots(8, 4, sharex=True, figsize=(10, 15))
            for row in range(0, 8):
                for col in range(0, 4):
                    count = 32 * page + 4 * row + col
                    if count < num_obs_plotted:
                        axes[row, col].scatter(group['increment'], group[obs_plotted[count]])
                        axes[row, col].plot(group['increment'], group[obs_plotted[count]], 'r')
                        axes[row, col].set_title(obs_plotted[count])
                        axes[row, col].set_xticks([1, 2, 3, 4, 5])
                        axes[row, col].tick_params(direction='in')
                        if row == 3:
                            axes[row, col].set_xlabel('Increment')
            plt.tight_layout()

            if filetype is None:
                plt.show()
            else:
                plt.savefig(os.path.join(figures_dir, "{0}_jactest_{1}.{2}".format(param, page, filetype)))
            plt.close()



""" pyemu utils module contains lots of useful functions and
classes, including support for geostatistical interpolation and
covariance matrices, pilot point setup and processing and
functionality dedicated to wrapping MODFLOW models into
the PEST(++) model independent framework
"""
from .helpers import *
from .geostats import *
from .pp_utils import *
from .gw_utils import *
from .os_utils import *
from .smp_utils import *




"""Geostatistical analyses within the pyemu framework.
Support for Ordinary Kriging as well as construction of
covariance matrices from (nested) geostistical structures.
Also support for reading GSLIB and SGEMS files
"""
from __future__ import print_function
import os
import copy
from datetime import datetime
import multiprocessing as mp
import warnings
import numpy as np
import scipy.sparse
import pandas as pd
from pyemu.mat.mat_handler import Cov,SparseMatrix
from pyemu.utils.pp_utils import pp_file_to_dataframe
from ..pyemu_warnings import PyemuWarning

EPSILON = 1.0e-7

# class KrigeFactors(pd.DataFrame):
#     def __init__(self,*args,**kwargs):
#         super(KrigeFactors,self).__init__(*args,**kwargs)
#
#     def to_factors(self,filename,nrow,ncol,
#                    points_file="points.junk",
#                    zone_file="zone.junk"):
#         with open(filename,'w') as f:
#             f.write(points_file+'\n')
#             f.write(zone_file+'\n')
#             f.write("{0} {1}\n".format(ncol,nrow))
#             f.write("{0}\n".format(self.shape[0]))
#
#
#
#     def from_factors(self,filename):
#         raise NotImplementedError()



class GeoStruct(object):
    """a geostatistical structure object.  The object contains
    variograms and (optionally) nugget information.

    Parameters
    ----------
    nugget : (float)
        nugget contribution
    variograms : (list or Vario2d instance)
        variogram(s) associated with this GeoStruct instance
    name : (str)
        name to assign the structure.  Default is "struct1".
    transform : (str)
        the transformation to apply to the GeoStruct.  Can be
        "none" or "log", depending on the transformation of the
        property being represented by the GeoStruct.
        Default is "none

    Attributes
    ----------
    variograms : list
        the Vario2d objects associated with the GeoStruct
    name : str
        the name of the GeoStruct
    transform : str
        the transform of the GeoStruct

    Example
    -------
    ``>>>import pyemu``

    ``>>>v = pyemu.utils.geostats.ExpVario(a=1000,contribution=1.0)``

    ``>>>gs = pyemu.utils.geostats.GeoStruct(variograms=v,nugget=0.5)``

    """
    def __init__(self,nugget=0.0,variograms=[],name="struct1",
                 transform="none"):
        self.name = name
        self.nugget = float(nugget)
        if not isinstance(variograms,list):
            variograms = [variograms]
        for vario in variograms:
            assert isinstance(vario,Vario2d)
        self.variograms = variograms
        transform = transform.lower()
        assert transform in ["none","log"]
        self.transform = transform


    def __lt__(self,other):
        return self.name < other.name


    def __gt__(self,other):
        return self.name > other.name

    def to_struct_file(self, f):
        """ write a PEST-style structure file

        Parameters
        ----------
        f : (str or file handle)
            file to write the GeoStruct information to

        """
        if isinstance(f, str):
            f = open(f,'w')
        f.write("STRUCTURE {0}\n".format(self.name))
        f.write("  NUGGET {0}\n".format(self.nugget))
        f.write("  NUMVARIOGRAM {0}\n".format(len(self.variograms)))
        for v in self.variograms:
            f.write("  VARIOGRAM {0} {1}\n".format(v.name,v.contribution))
        f.write("  TRANSFORM {0}\n".format(self.transform))
        f.write("END STRUCTURE\n\n")
        for v in self.variograms:
            v.to_struct_file(f)

    def sparse_covariance_matrix(self,x,y,names):
        """build a pyemu.Cov instance from GeoStruct

                Parameters
                ----------
                x : (iterable of floats)
                    x-coordinate locations
                y : (iterable of floats)
                    y-coordinate locations
                names : (iterable of str)
                   (parameter) names of locations.

                Returns
                -------
                sparse : pyemu.SparseMatrix
                    the sparse covariance matrix implied by this GeoStruct for the x,y pairs.

                Example
                -------
                ``>>>pp_df = pyemu.pp_utils.pp_file_to_dataframe("hkpp.dat")``

                ``>>>cov = gs.covariance_matrix(pp_df.x,pp_df.y,pp_df.name)``


                """

        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if not isinstance(y, np.ndarray):
            y = np.array(y)
        assert x.shape[0] == y.shape[0]
        assert x.shape[0] == len(names)

        iidx = [i for i in range(len(names))]
        jidx = list(iidx)
        data = list(np.zeros(x.shape[0])+self.nugget)

        for v in self.variograms:
            v.add_sparse_covariance_matrix(x,y,names,iidx,jidx,data)
        coo = scipy.sparse.coo_matrix((data,(iidx,jidx)),shape=(len(names),len(names)))
        coo.eliminate_zeros()
        coo.sum_duplicates()
        return SparseMatrix(coo,row_names=names,col_names=names)


    def covariance_matrix(self,x,y,names=None,cov=None):
        """build a pyemu.Cov instance from GeoStruct

        Parameters
        ----------
        x : (iterable of floats)
            x-coordinate locations
        y : (iterable of floats)
            y-coordinate locations
        names : (iterable of str)
            names of location. If None, cov must not be None.  Default is None.
        cov : (pyemu.Cov) instance
            an existing Cov instance.  The contribution of this GeoStruct is added
            to cov.  If cov is None, names must not be None. Default is None

        Returns
        -------
        cov : pyemu.Cov
            the covariance matrix implied by this GeoStruct for the x,y pairs.
            cov has row and column names supplied by the names argument unless
            the "cov" argument was passed.

        Note
        ----
        either "names" or "cov" must be passed.  If "cov" is passed, cov.shape
        must equal len(x) and len(y).

        Example
        -------
        ``>>>pp_df = pyemu.pp_utils.pp_file_to_dataframe("hkpp.dat")``

        ``>>>cov = gs.covariance_matrix(pp_df.x,pp_df.y,pp_df.name)``


        """

        if not isinstance(x,np.ndarray):
            x = np.array(x)
        if not isinstance(y,np.ndarray):
            y = np.array(y)
        assert x.shape[0] == y.shape[0]

        if names is not None:
            assert x.shape[0] == len(names)
            c = np.zeros((len(names),len(names)))
            np.fill_diagonal(c,self.nugget)
            cov = Cov(x=c,names=names)
        elif cov is not None:
            assert cov.shape[0] == x.shape[0]
            names = cov.row_names
            c = np.zeros((len(names),1))
            c += self.nugget
            cont = Cov(x=c,names=names,isdiagonal=True)
            cov += cont

        else:
            raise Exception("GeoStruct.covariance_matrix() requires either " +
                            "names or cov arg")
        for v in self.variograms:
            v.covariance_matrix(x,y,cov=cov)
        return cov

    def covariance(self,pt0,pt1):
        """get the covariance between two points implied by the GeoStruct.
        This is used during the ordinary kriging process to get the RHS

        Parameters
        ----------
        pt0 : (iterable length 2 of floats)
        pt1 : (iterable length 2 of floats)

        Returns
        -------
        covariance : float
            the covariance between pt0 and pt1 implied by the GeoStruct

        """
        #raise Exception()
        cov = self.nugget
        for vario in self.variograms:
            cov += vario.covariance(pt0,pt1)
        return cov

    def covariance_points(self,x0,y0,xother,yother):
        """ Get the covariance between point x0,y0 and the points
        contained in xother, yother.

        Parameters
        ----------
        x0 : (float)
            x-coordinate
        y0 : (float)
            y-coordinate
        xother : (iterable of floats)
            x-coordinate of other points
        yother : (iterable of floats)
            y-coordinate of other points

        Returns
        -------
        cov : numpy.ndarray
            a 1-D array of covariance between point x0,y0 and the
            points contained in xother, yother.  len(cov) = len(xother) =
            len(yother)

        """

        cov = np.zeros((len(xother))) + self.nugget
        for v in self.variograms:
            cov += v.covariance_points(x0,y0,xother,yother)
        return cov

    @property
    def sill(self):
        """ get the sill of the GeoStruct

        Return
        ------
        sill : float
            the sill of the (nested) GeoStruct, including nugget and contribution
            from each variogram
        """
        sill = self.nugget
        for v in self.variograms:
            sill += v.contribution
        return sill


    def plot(self,**kwargs):
        """ make a cheap plot of the GeoStruct

        Parameters
        ----------
        **kwargs : (dict)
            keyword arguments to use for plotting.

        Returns
        -------
        ax : matplotlib.pyplot.axis
            the axis with the GeoStruct plot

        Note
        ----
        optional arguments include "ax" (an existing axis),
        "individuals" (plot each variogram on a separate axis),
        "legend" (add a legend to the plot(s)).  All other kwargs
        are passed to matplotlib.pyplot.plot()

        """
        #
        if "ax" in kwargs:
            ax = kwargs.pop("ax")
        else:
            try:
                import matplotlib.pyplot as plt
            except Exception as e:
                raise Exception("error importing matplotlib: {0}".format(str(e)))

            ax = plt.subplot(111)
        legend = kwargs.pop("legend",False)
        individuals = kwargs.pop("individuals",False)
        xmx = max([v.a*3.0 for v in self.variograms])
        x = np.linspace(0,xmx,100)
        y = np.zeros_like(x)
        for v in self.variograms:
            yv = v.inv_h(x)
            if individuals:
                ax.plot(x,yv,label=v.name,**kwargs)
            y += yv
        y += self.nugget
        ax.plot(x,y,label=self.name,**kwargs)
        if legend:
            ax.legend()
        ax.set_xlabel("distance")
        ax.set_ylabel("$\gamma$")
        return ax

    def __str__(self):
        """ the str representation of the GeoStruct

        Returns
        -------
        str : str
            the str representation of the GeoStruct
        """
        s = ''
        s += 'name:{0},nugget:{1},structures:\n'.format(self.name,self.nugget)
        for v in self.variograms:
            s += str(v)
        return s


# class LinearUniversalKrige(object):
#     def __init__(self,geostruct,point_data):
#         if isinstance(geostruct,str):
#             geostruct = read_struct_file(geostruct)
#         assert isinstance(geostruct,GeoStruct),"need a GeoStruct, not {0}".\
#             format(type(geostruct))
#         self.geostruct = geostruct
#         if isinstance(point_data,str):
#             point_data = pp_file_to_dataframe(point_data)
#         assert isinstance(point_data,pd.DataFrame)
#         assert 'name' in point_data.columns,"point_data missing 'name'"
#         assert 'x' in point_data.columns, "point_data missing 'x'"
#         assert 'y' in point_data.columns, "point_data missing 'y'"
#         assert "value" in point_data.columns,"point_data missing 'value'"
#         self.point_data = point_data
#         self.point_data.index = self.point_data.name
#         self.interp_data = None
#         self.spatial_reference = None
#         #X, Y = np.meshgrid(point_data.x,point_data.y)
#         #self.point_data_dist = pd.DataFrame(data=np.sqrt((X - X.T) ** 2 + (Y - Y.T) ** 2),
#         #                                    index=point_data.name,columns=point_data.name)
#         self.point_cov_df = self.geostruct.covariance_matrix(point_data.x,
#                                                             point_data.y,
#                                                             point_data.name).to_dataframe()
#         #for name in self.point_cov_df.index:
#         #    self.point_cov_df.loc[name,name] -= self.geostruct.nugget
#
#
#     def estimate_grid(self,spatial_reference,zone_array=None,minpts_interp=1,
#                           maxpts_interp=20,search_radius=1.0e+10,verbose=False,
#                           var_filename=None):
#
#         self.spatial_reference = spatial_reference
#         self.interp_data = None
#         #assert isinstance(spatial_reference,SpatialReference)
#         try:
#             x = self.spatial_reference.xcentergrid.copy()
#             y = self.spatial_reference.ycentergrid.copy()
#         except Exception as e:
#             raise Exception("spatial_reference does not have proper attributes:{0}"\
#                             .format(str(e)))
#
#         if var_filename is not None:
#             arr = np.zeros((self.spatial_reference.nrow,
#                             self.spatial_reference.ncol)) - 1.0e+30
#
#         df = self.estimate(x.ravel(),y.ravel(),
#                            minpts_interp=minpts_interp,
#                            maxpts_interp=maxpts_interp,
#                            search_radius=search_radius,
#                            verbose=verbose)
#         if var_filename is not None:
#             arr = df.err_var.values.reshape(x.shape)
#             np.savetxt(var_filename,arr,fmt="%15.6E")
#         arr = df.estimate.values.reshape(x.shape)
#         return arr
#
#
#     def estimate(self,x,y,minpts_interp=1,maxpts_interp=20,
#                      search_radius=1.0e+10,verbose=False):
#         assert len(x) == len(y)
#
#         # find the point data to use for each interp point
#         sqradius = search_radius**2
#         df = pd.DataFrame(data={'x':x,'y':y})
#         inames,idist,ifacts,err_var = [],[],[],[]
#         estimates = []
#         sill = self.geostruct.sill
#         pt_data = self.point_data
#         ptx_array = pt_data.x.values
#         pty_array = pt_data.y.values
#         ptnames = pt_data.name.values
#         #if verbose:
#         print("starting interp point loop for {0} points".format(df.shape[0]))
#         start_loop = datetime.now()
#         for idx,(ix,iy) in enumerate(zip(df.x,df.y)):
#             if np.isnan(ix) or np.isnan(iy): #if nans, skip
#                 inames.append([])
#                 idist.append([])
#                 ifacts.append([])
#                 err_var.append(np.NaN)
#                 continue
#             if verbose:
#                 istart = datetime.now()
#                 print("processing interp point:{0} of {1}".format(idx,df.shape[0]))
#             # if verbose == 2:
#             #     start = datetime.now()
#             #     print("calc ipoint dist...",end='')
#
#             #  calc dist from this interp point to all point data...slow
#             dist = pd.Series((ptx_array-ix)**2 + (pty_array-iy)**2,ptnames)
#             dist.sort_values(inplace=True)
#             dist = dist.loc[dist <= sqradius]
#
#             # if too few points were found, skip
#             if len(dist) < minpts_interp:
#                 inames.append([])
#                 idist.append([])
#                 ifacts.append([])
#                 err_var.append(sill)
#                 estimates.append(np.NaN)
#                 continue
#
#             # only the maxpts_interp points
#             dist = dist.iloc[:maxpts_interp].apply(np.sqrt)
#             pt_names = dist.index.values
#             # if one of the points is super close, just use it and skip
#             if dist.min() <= EPSILON:
#                 ifacts.append([1.0])
#                 idist.append([EPSILON])
#                 inames.append([dist.idxmin()])
#                 err_var.append(self.geostruct.nugget)
#                 estimates.append(self.point_data.loc[dist.idxmin(),"value"])
#                 continue
#             # if verbose == 2:
#             #     td = (datetime.now()-start).total_seconds()
#             #     print("...took {0}".format(td))
#             #     start = datetime.now()
#             #     print("extracting pt cov...",end='')
#
#             #vextract the point-to-point covariance matrix
#             point_cov = self.point_cov_df.loc[pt_names,pt_names]
#             # if verbose == 2:
#             #     td = (datetime.now()-start).total_seconds()
#             #     print("...took {0}".format(td))
#             #     print("forming ipt-to-point cov...",end='')
#
#             # calc the interp point to points covariance
#             ptx = self.point_data.loc[pt_names,"x"]
#             pty = self.point_data.loc[pt_names,"y"]
#             interp_cov = self.geostruct.covariance_points(ix,iy,ptx,pty)
#
#             if verbose == 2:
#                 td = (datetime.now()-start).total_seconds()
#                 print("...took {0}".format(td))
#                 print("forming lin alg components...",end='')
#
#             # form the linear algebra parts and solve
#             d = len(pt_names) + 3 # +1 for lagrange mult + 2 for x and y coords
#             npts = len(pt_names)
#             A = np.ones((d,d))
#             A[:npts,:npts] = point_cov.values
#             A[npts,npts] = 0.0 #unbiaised constraint
#             A[-2,:npts] = ptx #x coords for linear trend
#             A[:npts,-2] = ptx
#             A[-1,:npts] = pty #y coords for linear trend
#             A[:npts,-1] = pty
#             A[npts:,npts:] = 0
#             print(A)
#             rhs = np.ones((d,1))
#             rhs[:npts,0] = interp_cov
#             rhs[-2,0] = ix
#             rhs[-1,0] = iy
#             # if verbose == 2:
#             #     td = (datetime.now()-start).total_seconds()
#             #     print("...took {0}".format(td))
#             #     print("solving...",end='')
#             # # solve
#             facs = np.linalg.solve(A,rhs)
#             assert len(facs) - 3 == len(dist)
#             estimate = facs[-3] + (ix * facs[-2]) + (iy * facs[-1])
#             estimates.append(estimate[0])
#             err_var.append(float(sill + facs[-1] - sum([f*c for f,c in zip(facs[:-1],interp_cov)])))
#             inames.append(pt_names)
#
#             idist.append(dist.values)
#             ifacts.append(facs[:-1,0])
#             # if verbose == 2:
#             #     td = (datetime.now()-start).total_seconds()
#             #     print("...took {0}".format(td))
#             if verbose:
#                 td = (datetime.now()-istart).total_seconds()
#                 print("point took {0}".format(td))
#         df["idist"] = idist
#         df["inames"] = inames
#         df["ifacts"] = ifacts
#         df["err_var"] = err_var
#         df["estimate"] = estimates
#         self.interp_data = df
#         td = (datetime.now() - start_loop).total_seconds()
#         print("took {0}".format(td))
#         return df


class OrdinaryKrige(object):
    """ Ordinary Kriging using Pandas and Numpy.

    Parameters
    ----------
    geostruct : (GeoStruct)
        a pyemu.geostats.GeoStruct to use for the kriging
    point_data : (pandas.DataFrame)
        the conditioning points to use for kriging.  point_data must contain
        columns "name", "x", "y".

    Note
    ----
    if point_data is an str, then it is assumed to be a pilot points file
    and is loaded as such using pyemu.pp_utils.pp_file_to_dataframe()

    If zoned interpolation is used for grid-based interpolation, then
    point_data must also contain a "zone" column


    Example
    -------
    ``>>>import pyemu``

    ``>>>v = pyemu.utils.geostats.ExpVario(a=1000,contribution=1.0)``

    ``>>>gs = pyemu.utils.geostats.GeoStruct(variograms=v,nugget=0.5)``

    ``>>>pp_df = pyemu.pp_utils.pp_file_to_dataframe("hkpp.dat")``

    ``>>>ok = pyemu.utils.geostats.OrdinaryKrige(gs,pp_df)``
    """

    def __init__(self,geostruct,point_data):
        if isinstance(geostruct,str):
            geostruct = read_struct_file(geostruct)
        assert isinstance(geostruct,GeoStruct),"need a GeoStruct, not {0}".\
            format(type(geostruct))
        self.geostruct = geostruct
        if isinstance(point_data,str):
            point_data = pp_file_to_dataframe(point_data)
        assert isinstance(point_data,pd.DataFrame)
        assert 'name' in point_data.columns,"point_data missing 'name'"
        assert 'x' in point_data.columns, "point_data missing 'x'"
        assert 'y' in point_data.columns, "point_data missing 'y'"
        #check for duplicates in point data
        unique_name = point_data.name.unique()
        if len(unique_name) != point_data.shape[0]:
            warnings.warn("duplicates detected in point_data..attempting to rectify",PyemuWarning)
            ux_std = point_data.groupby(point_data.name).std()['x']
            if ux_std.max() > 0.0:
                raise Exception("duplicate point_info entries with name {0} have different x values"
                                .format(uname))
            uy_std = point_data.groupby(point_data.name).std()['y']
            if uy_std.max() > 0.0:
                raise Exception("duplicate point_info entries with name {0} have different y values"
                                .format(uname))

            self.point_data = point_data.drop_duplicates(subset=["name"])
        else:
            self.point_data = point_data.copy()
        self.point_data.index = self.point_data.name
        self.check_point_data_dist()
        self.interp_data = None
        self.spatial_reference = None
        #X, Y = np.meshgrid(point_data.x,point_data.y)
        #self.point_data_dist = pd.DataFrame(data=np.sqrt((X - X.T) ** 2 + (Y - Y.T) ** 2),
        #                                    index=point_data.name,columns=point_data.name)
        self.point_cov_df = self.geostruct.covariance_matrix(self.point_data.x,
                                                            self.point_data.y,
                                                            self.point_data.name).to_dataframe()
        #for name in self.point_cov_df.index:
        #    self.point_cov_df.loc[name,name] -= self.geostruct.nugget

    def check_point_data_dist(self, rectify=False):
        """ check for point_data entries that are closer than
        EPSILON distance - this will cause a singular kriging matrix.

        Parameters
        ----------
        rectify : (boolean)
            flag to fix the problems with point_data
            by dropping additional points that are
            closer than EPSILON distance.  Default is False

        Note
        ----
        this method will issue warnings for points that are closer
        than EPSILON distance

        """

        ptx_array = self.point_data.x.values
        pty_array = self.point_data.y.values
        ptnames = self.point_data.name.values
        drop = []
        for i in range(self.point_data.shape[0]):
            ix,iy,iname = ptx_array[i],pty_array[i],ptnames[i]
            dist = pd.Series((ptx_array[i+1:] - ix) ** 2 + (pty_array[i+1:] - iy) ** 2, ptnames[i+1:])
            if dist.min() < EPSILON**2:
                print(iname,ix,iy)
                warnings.warn("points {0} and {1} are too close. This will cause a singular kriging matrix ".\
                              format(iname,dist.idxmin()),PyemuWarning)
                drop_idxs = dist.loc[dist<=EPSILON**2]
                drop.extend([pt for pt in list(drop_idxs.index) if pt not in drop])
        if rectify and len(drop) > 0:
            print("rectifying point data by removing the following points: {0}".format(','.join(drop)))
            print(self.point_data.shape)
            self.point_data = self.point_data.loc[self.point_data.index.map(lambda x: x not in drop),:]
            print(self.point_data.shape)

    #def prep_for_ppk2fac(self,struct_file="structure.dat",pp_file="points.dat",):
    #    pass



    def calc_factors_grid(self,spatial_reference,zone_array=None,minpts_interp=1,
                          maxpts_interp=20,search_radius=1.0e+10,verbose=False,
                          var_filename=None, forgive=False):
        """ calculate kriging factors (weights) for a structured grid.

        Parameters
        ----------
        spatial_reference : (flopy.utils.reference.SpatialReference)
            a spatial reference that describes the orientation and
            spatail projection of the the structured grid
        zone_array : (numpy.ndarray)
            an integer array of zones to use for kriging.  If not None,
            then point_data must also contain a "zone" column.  point_data
            entries with a zone value not found in zone_array will be skipped.
            If None, then all point_data will (potentially) be used for
            interpolating each grid node. Default is None
        minpts_interp : (int)
            minimum number of point_data entires to use for interpolation at
            a given grid node.  grid nodes with less than minpts_interp
            point_data found will be skipped (assigned np.NaN).  Defaut is 1
        maxpts_interp : (int)
            maximum number of point_data entries to use for interpolation at
            a given grid node.  A larger maxpts_interp will yield "smoother"
            interplation, but using a large maxpts_interp will slow the
            (already) slow kriging solution process and may lead to
            memory errors. Default is 20.
        search_radius : (float)
            the size of the region around a given grid node to search for
            point_data entries. Default is 1.0e+10
        verbose : (boolean)
            a flag to  echo process to stdout during the interpolatino process.
            Default is False
        var_filename : (str)
            a filename to save the kriging variance for each interpolated grid node.
            Default is None.
        forgive : (boolean)
            flag to continue if inversion of the kriging matrix failes at one or more
            grid nodes.  Inversion usually fails if the kriging matrix is singular,
            resulting from point_data entries closer than EPSILON distance.  If True,
            warnings are issued for each failed inversion.  If False, an exception
            is raised for failed matrix inversion.

        Returns
        -------
        df : pandas.DataFrame
            a dataframe with information summarizing the ordinary kriging
            process for each grid node

        Note
        ----
        this method calls OrdinaryKrige.calc_factors()


        Example
        -------
        ``>>>import flopy``

        ``>>>import pyemu``

        ``>>>v = pyemu.utils.geostats.ExpVario(a=1000,contribution=1.0)``

        ``>>>gs = pyemu.utils.geostats.GeoStruct(variograms=v,nugget=0.5)``

        ``>>>pp_df = pyemu.pp_utils.pp_file_to_dataframe("hkpp.dat")``

        ``>>>ok = pyemu.utils.geostats.OrdinaryKrige(gs,pp_df)``

        ``>>>m = flopy.modflow.Modflow.load("mymodel.nam")``

        ``>>>df = ok.calc_factors_grid(m.sr,zone_array=m.bas6.ibound[0].array,``

        ``>>>                          var_filename="ok_var.dat")``

        ``>>>ok.to_grid_factor_file("factors.dat")``

        """

        self.spatial_reference = spatial_reference
        self.interp_data = None
        #assert isinstance(spatial_reference,SpatialReference)
        try:
            x = self.spatial_reference.xcentergrid.copy()
            y = self.spatial_reference.ycentergrid.copy()
        except Exception as e:
            raise Exception("spatial_reference does not have proper attributes:{0}"\
                            .format(str(e)))

        if var_filename is not None:
                arr = np.zeros((self.spatial_reference.nrow,
                                self.spatial_reference.ncol)) - 1.0e+30

        # the simple case of no zone array: ignore point_data zones
        if zone_array is None:
            df = self.calc_factors(x.ravel(),y.ravel(),
                               minpts_interp=minpts_interp,
                               maxpts_interp=maxpts_interp,
                               search_radius=search_radius,
                               verbose=verbose, forgive=forgive)
            if var_filename is not None:
                arr = df.err_var.values.reshape(x.shape)
                np.savetxt(var_filename,arr,fmt="%15.6E")

        if zone_array is not None:
            assert zone_array.shape == x.shape
            if "zone" not in self.point_data.columns:
                warnings.warn("'zone' columns not in point_data, assigning generic zone",PyemuWarning)
                self.point_data.loc[:,"zone"] = 1
            pt_data_zones = self.point_data.zone.unique()
            dfs = []
            for pt_data_zone in pt_data_zones:
                if pt_data_zone not in zone_array:
                    warnings.warn("pt zone {0} not in zone array {1}, skipping".\
                                  format(pt_data_zone,np.unique(zone_array)),PyemuWarning)
                    continue
                xzone,yzone = x.copy(),y.copy()
                xzone[zone_array!=pt_data_zone] = np.NaN
                yzone[zone_array!=pt_data_zone] = np.NaN
                df = self.calc_factors(xzone.ravel(),yzone.ravel(),
                                       minpts_interp=minpts_interp,
                                       maxpts_interp=maxpts_interp,
                                       search_radius=search_radius,
                                       verbose=verbose,pt_zone=pt_data_zone,
                                       forgive=forgive)
                dfs.append(df)
                if var_filename is not None:
                    a = df.err_var.values.reshape(x.shape)
                    na_idx = np.isfinite(a)
                    arr[na_idx] = a[na_idx]
            if self.interp_data is None or self.interp_data.dropna().shape[0] == 0:
                raise Exception("no interpolation took place...something is wrong")
            df = pd.concat(dfs)
        if var_filename is not None:
            np.savetxt(var_filename,arr,fmt="%15.6E")
        return df

    def calc_factors(self,x,y,minpts_interp=1,maxpts_interp=20,
                     search_radius=1.0e+10,verbose=False,
                     pt_zone=None,forgive=False):
        """ calculate ordinary kriging factors (weights) for the points
        represented by arguments x and y

        Parameters
        ----------
        x : (iterable of floats)
            x-coordinates to calculate kriging factors for
        y : (iterable of floats)
            y-coordinates to calculate kriging factors for
        minpts_interp : (int)
            minimum number of point_data entires to use for interpolation at
            a given x,y interplation point.  interpolation points with less
            than minpts_interp point_data found will be skipped
            (assigned np.NaN).  Defaut is 1
        maxpts_interp : (int)
            maximum number of point_data entries to use for interpolation at
            a given x,y interpolation point.  A larger maxpts_interp will
            yield "smoother" interplation, but using a large maxpts_interp
            will slow the (already) slow kriging solution process and may
            lead to memory errors. Default is 20.
        search_radius : (float)
            the size of the region around a given x,y interpolation point to search for
            point_data entries. Default is 1.0e+10
        verbose : (boolean)
            a flag to  echo process to stdout during the interpolatino process.
            Default is False
        forgive : (boolean)
            flag to continue if inversion of the kriging matrix failes at one or more
            interpolation points.  Inversion usually fails if the kriging matrix is singular,
            resulting from point_data entries closer than EPSILON distance.  If True,
            warnings are issued for each failed inversion.  If False, an exception
            is raised for failed matrix inversion.

        Returns
        -------
        df : pandas.DataFrame
            a dataframe with information summarizing the ordinary kriging
            process for each interpolation points


        """

        assert len(x) == len(y)

        # find the point data to use for each interp point
        sqradius = search_radius**2
        df = pd.DataFrame(data={'x':x,'y':y})
        inames,idist,ifacts,err_var = [],[],[],[]
        sill = self.geostruct.sill
        if pt_zone is None:
            ptx_array = self.point_data.x.values
            pty_array = self.point_data.y.values
            ptnames = self.point_data.name.values
        else:
            pt_data = self.point_data
            ptx_array = pt_data.loc[pt_data.zone==pt_zone,"x"].values
            pty_array = pt_data.loc[pt_data.zone==pt_zone,"y"].values
            ptnames = pt_data.loc[pt_data.zone==pt_zone,"name"].values
        #if verbose:
        print("starting interp point loop for {0} points".format(df.shape[0]))
        start_loop = datetime.now()
        for idx,(ix,iy) in enumerate(zip(df.x,df.y)):
            if np.isnan(ix) or np.isnan(iy): #if nans, skip
                inames.append([])
                idist.append([])
                ifacts.append([])
                err_var.append(np.NaN)
                continue
            if verbose:
                istart = datetime.now()
                print("processing interp point:{0} of {1}".format(idx,df.shape[0]))
            # if verbose == 2:
            #     start = datetime.now()
            #     print("calc ipoint dist...",end='')

            #  calc dist from this interp point to all point data...slow
            dist = pd.Series((ptx_array-ix)**2 + (pty_array-iy)**2,ptnames)
            dist.sort_values(inplace=True)
            dist = dist.loc[dist <= sqradius]

            # if too few points were found, skip
            if len(dist) < minpts_interp:
                inames.append([])
                idist.append([])
                ifacts.append([])
                err_var.append(sill)
                continue

            # only the maxpts_interp points
            dist = dist.iloc[:maxpts_interp].apply(np.sqrt)
            pt_names = dist.index.values
            # if one of the points is super close, just use it and skip
            if dist.min() <= EPSILON:
                ifacts.append([1.0])
                idist.append([EPSILON])
                inames.append([dist.idxmin()])
                err_var.append(self.geostruct.nugget)
                continue
            # if verbose == 2:
            #     td = (datetime.now()-start).total_seconds()
            #     print("...took {0}".format(td))
            #     start = datetime.now()
            #     print("extracting pt cov...",end='')

            #vextract the point-to-point covariance matrix
            point_cov = self.point_cov_df.loc[pt_names,pt_names]
            # if verbose == 2:
            #     td = (datetime.now()-start).total_seconds()
            #     print("...took {0}".format(td))
            #     print("forming ipt-to-point cov...",end='')

            # calc the interp point to points covariance
            interp_cov = self.geostruct.covariance_points(ix,iy,self.point_data.loc[pt_names,"x"],
                                                          self.point_data.loc[pt_names,"y"])

            if verbose == 2:
                td = (datetime.now()-start).total_seconds()
                print("...took {0} seconds".format(td))
                print("forming lin alg components...",end='')

            # form the linear algebra parts and solve
            d = len(pt_names) + 1 # +1 for lagrange mult
            A = np.ones((d,d))
            A[:-1,:-1] = point_cov.values
            A[-1,-1] = 0.0 #unbiaised constraint
            rhs = np.ones((d,1))
            rhs[:-1,0] = interp_cov
            # if verbose == 2:
            #     td = (datetime.now()-start).total_seconds()
            #     print("...took {0}".format(td))
            #     print("solving...",end='')
            # # solve
            try:
                facs = np.linalg.solve(A,rhs)
            except Exception as e:
                print("error solving for factors: {0}".format(str(e)))
                print("point:",ix,iy)
                print("dist:",dist)
                print("A:", A)
                print("rhs:", rhs)
                if forgive:
                    inames.append([])
                    idist.append([])
                    ifacts.append([])
                    err_var.append(np.NaN)
                    continue
                else:
                    raise Exception("error solving for factors:{0}".format(str(e)))
            assert len(facs) - 1 == len(dist)

            err_var.append(float(sill + facs[-1] - sum([f*c for f,c in zip(facs[:-1],interp_cov)])))
            inames.append(pt_names)

            idist.append(dist.values)
            ifacts.append(facs[:-1,0])
            # if verbose == 2:
            #     td = (datetime.now()-start).total_seconds()
            #     print("...took {0}".format(td))
            if verbose:
                td = (datetime.now()-istart).total_seconds()
                print("point took {0}".format(td))
        df["idist"] = idist
        df["inames"] = inames
        df["ifacts"] = ifacts
        df["err_var"] = err_var
        if pt_zone is None:
            self.interp_data = df
        else:
            if self.interp_data is None:
                self.interp_data = df
            else:
                self.interp_data = self.interp_data.append(df)
        td = (datetime.now() - start_loop).total_seconds()
        print("took {0} seconds".format(td))
        return df

    def to_grid_factors_file(self, filename,points_file="points.junk",
                             zone_file="zone.junk"):
        """ write a grid-based PEST-style factors file.  This file can be used with
        the fac2real() method to write an interpolated structured array

        Parameters
        ----------
        filename : (str)
            factor filename
        points_file : (str)
            points filename to add to the header of the factors file.
            Not used by fac2real() method.  Default is "points.junk"
        zone_file : (str)
            zone filename to add to the header of the factors file.
            Not used by fac2real() method.  Default is "zone.junk"

        Note
        ----
        this method should be called after OrdinaryKirge.calc_factors_grid()

        """
        if self.interp_data is None:
            raise Exception("ok.interp_data is None, must call calc_factors_grid() first")
        if self.spatial_reference is None:
            raise Exception("ok.spatial_reference is None, must call calc_factors_grid() first")
        with open(filename, 'w') as f:
            f.write(points_file + '\n')
            f.write(zone_file + '\n')
            f.write("{0} {1}\n".format(self.spatial_reference.ncol, self.spatial_reference.nrow))
            f.write("{0}\n".format(self.point_data.shape[0]))
            [f.write("{0}\n".format(name)) for name in self.point_data.name]
            t = 0
            if self.geostruct.transform == "log":
                t = 1
            pt_names = list(self.point_data.name)
            for idx,names,facts in zip(self.interp_data.index,self.interp_data.inames,self.interp_data.ifacts):
                if len(facts) == 0:
                    continue
                n_idxs = [pt_names.index(name) for name in names]
                f.write("{0} {1} {2} {3:8.5e} ".format(idx+1, t, len(names), 0.0))
                [f.write("{0} {1:12.8g} ".format(i+1, w)) for i, w in zip(n_idxs, facts)]
                f.write("\n")


class Vario2d(object):
    """base class for 2-D variograms.

    Parameters
    ----------
    contribution : (float)
        sill of the variogram
    a : (float)
        (practical) range of correlation
    anisotropy : (float)
        Anisotropy ratio. Default is 1.0
    bearing : (float)
        angle in degrees East of North corresponding to anisotropy ellipse.
        Default is 0.0
    name : (str)
        name of the variogram.  Default is "var1"

    Returns
    -------
    Vario2d : Vario2d

    Note
    ----
    This base class should not be instantiated directly as it does not implement
    an h_function() method.

    """

    def __init__(self,contribution,a,anisotropy=1.0,bearing=0.0,name="var1"):
        self.name = name
        self.epsilon = EPSILON
        self.contribution = float(contribution)
        assert self.contribution > 0.0
        self.a = float(a)
        assert self.a > 0.0
        self.anisotropy = float(anisotropy)
        assert self.anisotropy > 0.0
        self.bearing = float(bearing)

    def to_struct_file(self, f):
        """ write the Vario2d to a PEST-style structure file

        Parameters
        ----------
        f : (str or file handle)
            item to write to

        """
        if isinstance(f, str):
            f = open(f,'w')
        f.write("VARIOGRAM {0}\n".format(self.name))
        f.write("  VARTYPE {0}\n".format(self.vartype))
        f.write("  A {0}\n".format(self.a))
        f.write("  ANISOTROPY {0}\n".format(self.anisotropy))
        f.write("  BEARING {0}\n".format(self.bearing))
        f.write("END VARIOGRAM\n\n")

    @property
    def bearing_rads(self):
        """ get the bearing of the Vario2d in radians

        Returns
        -------
        bearing_rads : float
            the Vario2d bearing in radians
        """
        return (np.pi / 180.0 ) * (90.0 - self.bearing)

    @property
    def rotation_coefs(self):
        """ get the rotation coefficents in radians

        Returns
        -------
        rotation_coefs : list
            the rotation coefficients implied by Vario2d.bearing


        """
        return [np.cos(self.bearing_rads),
                np.sin(self.bearing_rads),
                -1.0*np.sin(self.bearing_rads),
                np.cos(self.bearing_rads)]

    def inv_h(self,h):
        """ the inverse of the h_function.  Used for plotting

        Parameters
        ----------
        h : (float)
            the value of h_function to invert

        Returns
        -------
        inv_h : float
            the inverse of h

        """
        return self.contribution - self._h_function(h)

    def plot(self,**kwargs):
        """ get a cheap plot of the Vario2d

        Parameters
        ----------
        **kwargs : (dict)
            keyword arguments to use for plotting

        Returns
        -------
        ax : matplotlib.pyplot.axis

        Note
        ----
        optional arguments in kwargs include
        "ax" (existing matplotlib.pyplot.axis).  Other
        kwargs are passed to matplotlib.pyplot.plot()

        """
        try:
            import matplotlib.pyplot as plt
        except Exception as e:
            raise Exception("error importing matplotlib: {0}".format(str(e)))

        ax = kwargs.pop("ax",plt.subplot(111))
        x = np.linspace(0,self.a*3,100)
        y = self.inv_h(x)
        ax.set_xlabel("distance")
        ax.set_ylabel("$\gamma$")
        ax.plot(x,y,**kwargs)
        return ax


    def add_sparse_covariance_matrix(self,x,y,names,iidx,jidx,data):

        """build a pyemu.SparseMatrix instance implied by Vario2d

        Parameters
        ----------
        x : (iterable of floats)
            x-coordinate locations
        y : (iterable of floats)
            y-coordinate locations
        names : (iterable of str)
            names of locations. If None, cov must not be None
        iidx : 1-D ndarray
            i row indices
        jidx : 1-D ndarray
            j col indices
        data : 1-D ndarray
            nonzero entries


        Returns
        -------
        None

        """
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if not isinstance(y, np.ndarray):
            y = np.array(y)
        assert x.shape[0] == y.shape[0]


        assert x.shape[0] == len(names)
        #     c = np.zeros((len(names), len(names)))
        #     np.fill_diagonal(c, self.contribution)
        #     cov = Cov(x=c, names=names)
        # elif cov is not None:
        #     assert cov.shape[0] == x.shape[0]
        #     names = cov.row_names
        #     c = np.zeros((len(names), 1)) + self.contribution
        #     cont = Cov(x=c, names=names, isdiagonal=True)
        #     cov += cont
        #
        # else:
        #     raise Exception("Vario2d.covariance_matrix() requires either" +
        #                     "names or cov arg")
        # rc = self.rotation_coefs
        for i,name in enumerate(names):
            iidx.append(i)
            jidx.append(i)
            data.append(self.contribution)

        for i1, (n1, x1, y1) in enumerate(zip(names, x, y)):
            dx = x1 - x[i1 + 1:]
            dy = y1 - y[i1 + 1:]
            dxx, dyy = self._apply_rotation(dx, dy)
            h = np.sqrt(dxx * dxx + dyy * dyy)

            h[h < 0.0] = 0.0
            cv = self._h_function(h)
            if np.any(np.isnan(cv)):
                raise Exception("nans in cv for i1 {0}".format(i1))
            #cv[h>self.a] = 0.0
            j = list(np.arange(i1+1,x.shape[0]))
            i = [i1] * len(j)
            iidx.extend(i)
            jidx.extend(j)
            data.extend(list(cv))
            # replicate across the diagonal
            iidx.extend(j)
            jidx.extend(i)
            data.extend(list(cv))



    def covariance_matrix(self,x,y,names=None,cov=None):
        """build a pyemu.Cov instance implied by Vario2d

        Parameters
        ----------
        x : (iterable of floats)
            x-coordinate locations
        y : (iterable of floats)
            y-coordinate locations
        names : (iterable of str)
            names of locations. If None, cov must not be None
        cov : (pyemu.Cov)
            an existing Cov instance.  Vario2d contribution is added to cov

        Returns
        -------
        cov : pyemu.Cov

        Note
        ----
        either names or cov must not be None.

        """
        if not isinstance(x,np.ndarray):
            x = np.array(x)
        if not isinstance(y,np.ndarray):
            y = np.array(y)
        assert x.shape[0] == y.shape[0]

        if names is not None:
            assert x.shape[0] == len(names)
            c = np.zeros((len(names),len(names)))
            np.fill_diagonal(c,self.contribution)
            cov = Cov(x=c,names=names)
        elif cov is not None:
            assert cov.shape[0] == x.shape[0]
            names = cov.row_names
            c = np.zeros((len(names),1)) + self.contribution
            cont = Cov(x=c,names=names,isdiagonal=True)
            cov += cont

        else:
            raise Exception("Vario2d.covariance_matrix() requires either" +
                            "names or cov arg")
        rc = self.rotation_coefs
        for i1,(n1,x1,y1) in enumerate(zip(names,x,y)):
            dx = x1 - x[i1+1:]
            dy = y1 - y[i1+1:]
            dxx,dyy = self._apply_rotation(dx,dy)
            h = np.sqrt(dxx*dxx + dyy*dyy)

            h[h<0.0] = 0.0
            h = self._h_function(h)
            if np.any(np.isnan(h)):
                raise Exception("nans in h for i1 {0}".format(i1))
            cov.x[i1,i1+1:] += h
        for i in range(len(names)):
            cov.x[i+1:,i] = cov.x[i,i+1:]
        return cov

    def _apply_rotation(self,dx,dy):
        """ private method to rotate points
        according to Vario2d.bearing and Vario2d.anisotropy

        Parameters
        ----------
        dx : (float or numpy.ndarray)
            x-coordinates to rotate
        dy : (float or numpy.ndarray)
            y-coordinates to rotate

        Returns
        -------
            dxx : (float or numpy.ndarray)
                rotated x-coordinates
            dyy : (float or numpy.ndarray)
                rotated y-coordinates

        """
        if self.anisotropy == 1.0:
            return dx,dy
        rcoefs = self.rotation_coefs
        dxx = (dx * rcoefs[0]) +\
             (dy * rcoefs[1])
        dyy = ((dx * rcoefs[2]) +\
             (dy * rcoefs[3])) *\
             self.anisotropy
        return dxx,dyy

    def covariance_points(self,x0,y0,xother,yother):
        """ get the covariance between base point x0,y0 and
        other points xother,yother implied by Vario2d

        Parameters
        ----------
        x0 : (float)
            x-coordinate of base point
        y0 : (float)
            y-coordinate of base point
        xother : (float or numpy.ndarray)
            x-coordinates of other points
        yother : (float or numpy.ndarray)
            y-coordinates of other points

        Returns
        -------
        cov : numpy.ndarray
            covariance between base point and other points implied by
            Vario2d.

        Note
        ----
        len(cov) = len(xother) = len(yother)

        """
        dx = x0 - xother
        dy = y0 - yother
        dxx,dyy = self._apply_rotation(dx,dy)
        h = np.sqrt(dxx*dxx + dyy*dyy)
        return self._h_function(h)

    def covariance(self,pt0,pt1):
        """ get the covarince between two points implied by Vario2d

        Parameters
        ----------
        pt0 : (iterable of len 2)
            first point x and y
        pt1 : (iterable of len 2)
            second point x and y

        Returns
        -------
        cov : float
            covariance between pt0 and pt1

        """

        x = np.array([pt0[0],pt1[0]])
        y = np.array([pt0[1],pt1[1]])
        names = ["n1","n2"]
        return self.covariance_matrix(x,y,names=names).x[0,1]


    def __str__(self):
        """ get the str representation of Vario2d

        Returns
        -------
        str : str
        """
        s = "name:{0},contribution:{1},a:{2},anisotropy:{3},bearing:{4}\n".\
            format(self.name,self.contribution,self.a,\
                   self.anisotropy,self.bearing)
        return s

class ExpVario(Vario2d):
    """ Exponetial variogram derived type

    Parameters
    ----------
    contribution : (float)
        sill of the variogram
    a : (float)
        (practical) range of correlation
    anisotropy : (float)
        Anisotropy ratio. Default is 1.0
    bearing : (float)
        angle in degrees East of North corresponding to anisotropy ellipse.
        Default is 0.0
    name : (str)
        name of the variogram.  Default is "var1"

    Returns
    -------
    ExpVario : ExpVario

    Example
    -------
    ``>>>import pyemu``

    ``>>>v = pyemu.utils.geostats.ExpVario(a=1000,contribution=1.0)``

    """
    def __init__(self,contribution,a,anisotropy=1.0,bearing=0.0,name="var1"):
        super(ExpVario,self).__init__(contribution,a,anisotropy=anisotropy,
                                      bearing=bearing,name=name)
        self.vartype = 2

    def _h_function(self,h):
        """ private method exponential variogram "h" function

        Parameters
        ----------
        h : (float or numpy.ndarray)
            distance(s)

        Returns
        -------
        h_function : float or numpy.ndarray
            the value of the "h" function implied by the ExpVario

        """
        return self.contribution * np.exp(-1.0 * h / self.a)

class GauVario(Vario2d):
    """Gaussian variogram derived type

    Parameters
    ----------
    contribution : (float)
        sill of the variogram
    a : (float)
        (practical) range of correlation
    anisotropy : (float)
        Anisotropy ratio. Default is 1.0
    bearing : (float)
        angle in degrees East of North corresponding to anisotropy ellipse.
        Default is 0.0
    name : (str)
        name of the variogram.  Default is "var1"

    Returns
    -------
    GauVario : GauVario

    Note
    ----
    the Gaussian variogram can be unstable (not invertible) for long ranges.

    Example
    -------
    ``>>>import pyemu``

    ``>>>v = pyemu.utils.geostats.GauVario(a=1000,contribution=1.0)``

    """

    def __init__(self,contribution,a,anisotropy=1.0,bearing=0.0,name="var1"):
        super(GauVario,self).__init__(contribution,a,anisotropy=anisotropy,
                                      bearing=bearing,name=name)
        self.vartype = 3

    def _h_function(self,h):
        """ private method for the gaussian variogram "h" function

        Parameters
        ----------
        h : (float or numpy.ndarray)
            distance(s)

        Returns
        -------
        h_function : float or numpy.ndarray
            the value of the "h" function implied by the GauVario

        """

        hh = -1.0 * (h * h) / (self.a * self.a)
        return self.contribution * np.exp(hh)

class SphVario(Vario2d):
    """Spherical variogram derived type

    Parameters
    ----------
    contribution : (float)
        sill of the variogram
    a : (float)
        (practical) range of correlation
    anisotropy : (float)
        Anisotropy ratio. Default is 1.0
    bearing : (float)
        angle in degrees East of North corresponding to anisotropy ellipse.
        Default is 0.0
    name : (str)
        name of the variogram.  Default is "var1"

    Returns
    -------
    SphVario : SphVario

    Example
    -------
    ``>>>import pyemu``

    ``>>>v = pyemu.utils.geostats.SphVario(a=1000,contribution=1.0)``

    """

    def __init__(self,contribution,a,anisotropy=1.0,bearing=0.0,name="var1"):
        super(SphVario,self).__init__(contribution,a,anisotropy=anisotropy,
                                      bearing=bearing,name=name)
        self.vartype = 1

    def _h_function(self,h):
        """ private method for the spherical variogram "h" function

        Parameters
        ----------
        h : (float or numpy.ndarray)
            distance(s)

        Returns
        -------
        h_function : float or numpy.ndarray
            the value of the "h" function implied by the SphVario

        """

        hh = h / self.a
        h = self.contribution * (1.0 - (hh * (1.5 - (0.5 * hh * hh))))
        h[hh > 1.0] = 0.0
        return h
        # try:
        #     h[hh < 1.0] = 0.0
        #
        # except TypeError:
        #     if hh > 0.0:
        #         h = 0.0
        #return h
        # if hh < 1.0:
        #     return self.contribution * (1.0 - (hh * (1.5 - (0.5 * hh * hh))))
        # else:
        #     return 0.0





def read_struct_file(struct_file,return_type=GeoStruct):
    """read an existing PEST-type structure file into a GeoStruct instance

    Parameters
    ----------
    struct_file : (str)
        existing pest-type structure file
    return_type :  (object)
        the instance type to return.  Default is GeoStruct

    Returns
    -------
    GeoStruct : list or GeoStruct

    Note
    ----
    if only on structure is listed in struct_file, then return type
    is GeoStruct.  Otherwise, return type is a list of GeoStruct

    Example
    -------
    ``>>>import pyemu``

    ``>>>gs = pyemu.utils.geostats.reads_struct_file("struct.dat")``


    """

    VARTYPE = {1:SphVario,2:ExpVario,3:GauVario,4:None}
    assert os.path.exists(struct_file)
    structures = []
    variograms = []
    with open(struct_file,'r') as f:
        while True:
            line = f.readline()
            if line == '':
                break
            line = line.strip().lower()
            if line.startswith("structure"):
                name = line.strip().split()[1]
                nugget,transform,variogram_info = _read_structure_attributes(f)
                s = return_type(nugget=nugget,transform=transform,name=name)
                s.variogram_info = variogram_info
                # not sure what is going on, but if I don't copy s here,
                # all the structures end up sharing all the variograms later
                structures.append(copy.deepcopy(s))
            elif line.startswith("variogram"):
                name = line.strip().split()[1].lower()
                vartype,bearing,a,anisotropy = _read_variogram(f)
                if name in variogram_info:
                    v = VARTYPE[vartype](variogram_info[name],a,anisotropy=anisotropy,
                                         bearing=bearing,name=name)
                    variograms.append(v)

    for i,st in enumerate(structures):
        for vname in st.variogram_info:
            vfound = None
            for v in variograms:
                if v.name == vname:
                    vfound = v
                    break
            if vfound is None:
                raise Exception("variogram {0} not found for structure {1}".\
                                format(vname,s.name))

            st.variograms.append(vfound)
    if len(structures) == 1:
        return structures[0]
    return structures



def _read_variogram(f):
    """Function to instantiate a Vario2d from a PEST-style structure file

    Parameters
    ----------
    f : (file handle)
        file handle opened for reading

    Returns
    -------
    Vario2d : Vario2d
        Vario2d derived type

    """

    line = ''
    vartype = None
    bearing = 0.0
    a = None
    anisotropy = 1.0
    while "end variogram" not in line:
        line = f.readline()
        if line == '':
            raise Exception("EOF while read variogram")
        line = line.strip().lower().split()
        if line[0].startswith('#'):
            continue
        if line[0] == "vartype":
            vartype = int(line[1])
        elif line[0] == "bearing":
            bearing = float(line[1])
        elif line[0] == "a":
            a = float(line[1])
        elif line[0] == "anisotropy":
            anisotropy = float(line[1])
        elif line[0] == "end":
            break
        else:
            raise Exception("unrecognized arg in variogram:{0}".format(line[0]))
    return vartype,bearing,a,anisotropy


def _read_structure_attributes(f):
    """ function to read information from a PEST-style structure file

    Parameters
    ----------
    f : (file handle)
        file handle open for reading

    Returns
    -------
    nugget : float
        the GeoStruct nugget
    transform : str
        the GeoStruct transformation
    variogram_info : dict
        dictionary of structure-level variogram information

    """

    line = ''
    variogram_info = {}
    while "end structure" not in line:
        line = f.readline()
        if line == '':
            raise Exception("EOF while reading structure")
        line = line.strip().lower().split()
        if line[0].startswith('#'):
            continue
        if line[0] == "nugget":
            nugget = float(line[1])
        elif line[0] == "transform":
            transform = line[1]
        elif line[0] == "numvariogram":
            numvariograms = int(line[1])
        elif line[0] == "variogram":
            variogram_info[line[1]] = float(line[2])
        elif line[0] == "end":
            break
        elif line[0] == "mean":
            warnings.warn("'mean' attribute not supported, skipping",PyemuWarning)
        else:
            raise Exception("unrecognized line in structure definition:{0}".\
                            format(line[0]))
    assert numvariograms == len(variogram_info)
    return nugget,transform,variogram_info


def read_sgems_variogram_xml(xml_file,return_type=GeoStruct):
    """ function to read an SGEMS-type variogram XML file into
    a GeoStruct

    Parameters
    ----------
    xml_file : (str)
        SGEMS variogram XML file
    return_type :  (object)
        the instance type to return.  Default is GeoStruct

    Returns
    -------
    GeoStruct : GeoStruct


    Example
    -------
    ``>>>import pyemu``

    ``>>>gs = pyemu.utils.geostats.read_sgems_variogram_xml("sgems.xml")``

    """
    try:
        import xml.etree.ElementTree as ET

    except Exception as e:
        print("error import elementtree, skipping...")
    VARTYPE = {1: SphVario, 2: ExpVario, 3: GauVario, 4: None}
    assert os.path.exists(xml_file)
    tree = ET.parse(xml_file)
    gs_model = tree.getroot()
    structures = []
    variograms = []
    nugget = 0.0
    num_struct = 0
    for key,val in gs_model.items():
        #print(key,val)
        if str(key).lower() == "nugget":
            if len(val) > 0:
                nugget = float(val)
        if str(key).lower() == "structures_count":
            num_struct = int(val)
    if num_struct == 0:
        raise Exception("no structures found")
    if num_struct != 1:
        raise NotImplementedError()
    for structure in gs_model:
        vtype, contribution = None, None
        mx_range,mn_range = None, None
        x_angle,y_angle = None,None
        #struct_name = structure.tag
        for key,val in structure.items():
            key = str(key).lower()
            if key == "type":
                vtype = str(val).lower()
                if vtype.startswith("sph"):
                    vtype = SphVario
                elif vtype.startswith("exp"):
                    vtype = ExpVario
                elif vtype.startswith("gau"):
                    vtype = GauVario
                else:
                    raise Exception("unrecognized variogram type:{0}".format(vtype))

            elif key == "contribution":
                contribution = float(val)
            for item in structure:
                if item.tag.lower() == "ranges":
                    mx_range = float(item.attrib["max"])
                    mn_range = float(item.attrib["min"])
                elif item.tag.lower() == "angles":
                    x_angle = float(item.attrib["x"])
                    y_angle = float(item.attrib["y"])

        assert contribution is not None
        assert mn_range is not None
        assert mx_range is not None
        assert x_angle is not None
        assert y_angle is not None
        assert vtype is not None
        v = vtype(contribution=contribution,a=mx_range,
                  anisotropy=mx_range/mn_range,bearing=(180.0/np.pi)*np.arctan2(x_angle,y_angle),
                  name=structure.tag)
        return GeoStruct(nugget=nugget,variograms=[v])


def gslib_2_dataframe(filename,attr_name=None,x_idx=0,y_idx=1):
    """ function to read a GSLIB point data file into a pandas.DataFrame

    Parameters
    ----------
    filename : (str)
        GSLIB file
    attr_name : (str)
        the column name in the dataframe for the attribute.  If None, GSLIB file
        can have only 3 columns.  attr_name must be in the GSLIB file header
    x_idx : (int)
        the index of the x-coordinate information in the GSLIB file. Default is
        0 (first column)
    y_idx : (int)
        the index of the y-coordinate information in the GSLIB file.
        Default is 1 (second column)

    Returns
    -------
    df : pandas.DataFrame

    Raises
    ------
    exception if attr_name is None and GSLIB file has more than 3 columns

    Note
    ----
    assigns generic point names ("pt0, pt1, etc)

    Example
    -------
    ``>>>import pyemu``

    ``>>>df = pyemu.utiils.geostats.gslib_2_dataframe("prop.gslib",attr_name="hk")``


    """
    with open(filename,'r') as f:
        title = f.readline().strip()
        num_attrs = int(f.readline().strip())
        attrs = [f.readline().strip() for _ in range(num_attrs)]
        if attr_name is not None:
            assert attr_name in attrs,"{0} not in attrs:{1}".format(attr_name,','.join(attrs))
        else:
            assert len(attrs) == 3,"propname is None but more than 3 attrs in gslib file"
            attr_name = attrs[2]
        assert len(attrs) > x_idx
        assert len(attrs) > y_idx
        a_idx = attrs.index(attr_name)
        x,y,a = [],[],[]
        while True:
            line = f.readline()
            if line == '':
                break
            raw = line.strip().split()
            try:
                x.append(float(raw[x_idx]))
                y.append(float(raw[y_idx]))
                a.append(float(raw[a_idx]))
            except Exception as e:
                raise Exception("error paring line {0}: {1}".format(line,str(e)))
    df = pd.DataFrame({"x":x,"y":y,"value":a})
    df.loc[:,"name"] = ["pt{0}".format(i) for i in range(df.shape[0])]
    df.index = df.name
    return df


#class ExperimentalVariogram(object):
#    def __init__(self,na)

def load_sgems_exp_var(filename):
    """ read an SGEM experimental variogram into a sequence of
    pandas.DataFrames

    Parameters
    ----------
    filename : (str)
        an SGEMS experimental variogram XML file

    Returns
    -------
    dfs : list
        a list of pandas.DataFrames of x, y, pairs for each
        division in the experimental variogram

    """

    assert os.path.exists(filename)
    import xml.etree.ElementTree as etree
    tree = etree.parse(filename)
    root = tree.getroot()
    dfs = {}
    for variogram in root:
        #print(variogram.tag)
        for attrib in variogram:

            #print(attrib.tag,attrib.text)
            if attrib.tag == "title":
                title = attrib.text.split(',')[0].split('=')[-1]
            elif attrib.tag == "x":
                x = [float(i) for i in attrib.text.split()]
            elif attrib.tag == "y":
                y = [float(i) for i in attrib.text.split()]
            elif attrib.tag == "pairs":
                pairs = [int(i) for i in attrib.text.split()]

            for item in attrib:
                print(item,item.tag)
        df = pd.DataFrame({"x":x,"y":y,"pairs":pairs})
        df.loc[df.y<0.0,"y"] = np.NaN
        dfs[title] = df
    return dfs



def fac2real(pp_file=None,factors_file="factors.dat",out_file="test.ref",
             upper_lim=1.0e+30,lower_lim=-1.0e+30,fill_value=1.0e+30):
    """A python replication of the PEST fac2real utility for creating a
    structure grid array from previously calculated kriging factors (weights)

    Parameters
    ----------
    pp_file : (str)
        PEST-type pilot points file
    factors_file : (str)
        PEST-style factors file
    out_file : (str)
        filename of array to write.  If None, array is returned, else
        value of out_file is returned.  Default is "test.ref".
    upper_lim : (float)
        maximum interpolated value in the array.  Values greater than
        upper_lim are set to fill_value
    lower_lim : (float)
        minimum interpolated value in the array.  Values less than lower_lim
        are set to fill_value
    fill_value : (float)
        the value to assign array nodes that are not interpolated


    Returns
    -------
    arr : numpy.ndarray
        if out_file is None
    out_file : str
        if out_file it not None

    Example
    -------
    ``>>>import pyemu``

    ``>>>pyemu.utils.geostats.fac2real("hkpp.dat",out_file="hk_layer_1.ref")``

    """

    if pp_file is not None and isinstance(pp_file,str):
        assert os.path.exists(pp_file)
        # pp_data = pd.read_csv(pp_file,delim_whitespace=True,header=None,
        #                       names=["name","parval1"],usecols=[0,4])
        pp_data = pp_file_to_dataframe(pp_file)
        pp_data.loc[:,"name"] = pp_data.name.apply(lambda x: x.lower())
    elif pp_file is not None and isinstance(pp_file,pd.DataFrame):
        assert "name" in pp_file.columns
        assert "parval1" in pp_file.columns
        pp_data = pp_file
    else:
        raise Exception("unrecognized pp_file arg: must be str or pandas.DataFrame, not {0}"\
                        .format(type(pp_file)))
    assert os.path.exists(factors_file)
    f_fac = open(factors_file,'r')
    fpp_file = f_fac.readline()
    if pp_file is None and pp_data is None:
        pp_data = pp_file_to_dataframe(fpp_file)
        pp_data.loc[:, "name"] = pp_data.name.apply(lambda x: x.lower())

    fzone_file = f_fac.readline()
    ncol,nrow = [int(i) for i in f_fac.readline().strip().split()]
    npp = int(f_fac.readline().strip())
    pp_names = [f_fac.readline().strip().lower() for _ in range(npp)]

    # check that pp_names is sync'd with pp_data
    diff = set(list(pp_data.name)).symmetric_difference(set(pp_names))
    if len(diff) > 0:
        raise Exception("the following pilot point names are not common " +\
                        "between the factors file and the pilot points file " +\
                        ','.join(list(diff)))

    arr = np.zeros((nrow,ncol),dtype=np.float) + fill_value
    pp_dict = {int(name):val for name,val in zip(pp_data.index,pp_data.parval1)}
    try:
        pp_dict_log = {name:np.log10(val) for name,val in zip(pp_data.index,pp_data.parval1)}
    except:
        pp_dict_log = {}
    #for i in range(nrow):
    #    for j in range(ncol):
    while True:
        line = f_fac.readline()
        if len(line) == 0:
            #raise Exception("unexpected EOF in factors file")
            break
        try:
            inode,itrans,fac_data = parse_factor_line(line)
        except Exception as e:
            raise Exception("error parsing factor line {0}:{1}".format(line,str(e)))
        #fac_prods = [pp_data.loc[pp,"value"]*fac_data[pp] for pp in fac_data]
        if itrans == 0:
            fac_sum = sum([pp_dict[pp] * fac_data[pp] for pp in fac_data])
        else:
            fac_sum = sum([pp_dict_log[pp] * fac_data[pp] for pp in fac_data])
        if itrans != 0:
            fac_sum = 10**fac_sum
        #col = ((inode - 1) // nrow) + 1
        #row = inode - ((col - 1) * nrow)
        row = ((inode-1) // ncol) + 1
        col = inode - ((row - 1) * ncol)
        #arr[row-1,col-1] = np.sum(np.array(fac_prods))
        arr[row - 1, col - 1] = fac_sum
    arr[arr<lower_lim] = lower_lim
    arr[arr>upper_lim] = upper_lim

    #print(out_file,arr.min(),pp_data.parval1.min(),lower_lim)

    if out_file is not None:
        np.savetxt(out_file,arr,fmt="%15.6E",delimiter='')
        return out_file
    return arr

def parse_factor_line(line):
    """ function to parse a factor file line.  Used by fac2real()

    Parameters
    ----------
    line : (str)
        a factor line from a factor file

    Returns
    -------
    inode : int
        the inode of the grid node
    itrans : int
        flag for transformation of the grid node
    fac_data : dict
        a dictionary of point number, factor

    """

    raw = line.strip().split()
    inode,itrans,nfac = [int(i) for i in raw[:3]]
    fac_data = {int(raw[ifac])-1:float(raw[ifac+1]) for ifac in range(4,4+nfac*2,2)}
    # fac_data = {}
    # for ifac in range(4,4+nfac*2,2):
    #     pnum = int(raw[ifac]) - 1 #zero based to sync with pandas
    #     fac = float(raw[ifac+1])
    #     fac_data[pnum] = fac
    return inode,itrans,fac_data


"""pyEMU: python modules for Environmental Model Uncertainty analyses.  These
modules are designed to work directly and seamlessly with PEST and PEST++ model
independent interface.  pyEMU can also be used to setup this interface.

Several forms of uncertainty analyses are support including FOSM-based
analyses (pyemu.Schur and pyemu.ErrVar), Monte Carlo (including
GLUE and null-space Monte Carlo) and a prototype iterative Ensemble Smoother

"""

from .la import LinearAnalysis
from .sc import Schur
from .ev import ErrVar
from .en import Ensemble, ParameterEnsemble, ObservationEnsemble
from .mc import MonteCarlo
#from .inf import Influence
from .mat import Matrix, Jco, Cov, SparseMatrix
from .pst import Pst, pst_utils
from .utils import helpers, gw_utils, optimization,geostats, pp_utils, os_utils, smp_utils
from .plot import plot_utils
from .logger import Logger

from .prototypes import *

from ._version import get_versions
__version__ = get_versions()['version']
#del get_versions
